/*
 * ctrl_impl.c
 *
 *  Created on: 15.02.2016
 *      Author: AndreR
 */

#include "ctrl_impl.h"

#include <main/ctrl.h>
#include <main/skills.h>
#include "arm_math.h"
#include "log_msgs.h"
#include "constants.h"
#include "intercom1.h"
#include "motors.h"
#include "util/sys_time.h"
#include "spi2.h"
#include "main/network.h"
#include "util/init_hal.h"
#include "hal/buzzer.h"
#include "ui.h"
#include "util/intercom_reliable.h"
#include "util/arm_mat_util_f32.h"

CtrlImpl ctrlImpl;

BotParameters botParams = {
	.driveTrain = {
		.motor2WheelRatio = (15.0f/50.0f),
		.wheel2MotorRatio = (50.0f/15.0f),
		.motor = {
			.Km = 0.0255f,
			.Kn = (374.0f*2.0f*PI/60.0f),
			.Ke = (60.0f/(374.0f*2.0f*PI)),
			.Kf = 0.94e-4f,
			.R = 1.2f,
			.L = 560e-6f,
		}
	},
	.physical = { 0.0245f, 30.0f, 45.0f, 0.076f, 3.0f },
};

static const float out2MotX[129];
static const float out2MotY[129];
static const float out2MotW[129];
static const float mot2InX[129];
static const float mot2InY[129];

static void kdCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)dataLength;
	(void)pUserData;

	switch(pHeader->command)
	{
		case IC_KD_FEEDBACK:
		{
			KdFeedback* pFb = (KdFeedback*)pData;

			memcpy(&ctrlImpl.lastKdFeedback, pFb, sizeof(KdFeedback));
		}
		break;
	}
}

void CtrlImplInit()
{
	// register KD callback
	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &kdCallback;
	client.section = INTERCOM_SECTION_KD;
	IntercomAddClient(client);
}

void CtrlImplUpdateInput(CtrlInput* pInput, LogKickerDribbler* pKD)
{
	static uint32_t lastVisionTimestamp;
	static uint32_t lastGyrTimestamp;
	static uint32_t lastAccTimestamp;

	// motors
	pInput->enc.time = motors.mot[0].cur.velocityTimestamp;
	pInput->motorVol.time = motors.mot[0].cur.setVoltageTimestamp;

	for(uint8_t m = 0; m < 4; m++)
	{
		pInput->enc.vel[m] = motors.mot[m].cur.velocity * (2*PI);
		pInput->motorVol.vol[m] = motors.mot[m].cur.setVoltage;
	}

	// power
	pInput->power.voltage = ctrlImpl.lastKdFeedback.power.bat;
	pInput->power.current = ctrlImpl.lastKdFeedback.power.cur;
	pInput->power.batEmpty = ctrlImpl.lastKdFeedback.power.batEmpty;

	// IMU
	if(lastAccTimestamp != spi2.lastAccTime)
	{
		pInput->acc.linAcc[0] = spi2.acc[0];
		pInput->acc.linAcc[1] = spi2.acc[1];
		pInput->acc.linAcc[2] = spi2.acc[2];
		pInput->acc.time = spi2.lastAccTime;
		pInput->acc.updated = 1;

		lastAccTimestamp = spi2.lastAccTime;
	}

	if(lastGyrTimestamp != spi2.lastGyrTime)
	{
		pInput->gyr.rotVel[0] = spi2.gyr[0];
		pInput->gyr.rotVel[1] = spi2.gyr[1];
		pInput->gyr.rotVel[2] = spi2.gyr[2];
		pInput->gyr.time = spi2.lastGyrTime;
		pInput->gyr.updated = 1;

		lastGyrTimestamp = spi2.lastGyrTime;
	}

	// vision
	if(lastVisionTimestamp != network.lastVisionPosTimestamp)
	{
		pInput->vision.pos[0] = network.lastVisionPos[0]*0.001f;
		pInput->vision.pos[1] = network.lastVisionPos[1]*0.001f;
		pInput->vision.pos[2] = network.lastVisionPos[2]*0.001f;
		pInput->vision.time = network.lastVisionPosTimestamp;
		pInput->vision.delay = network.lastVisionDelay;
		pInput->vision.camId = network.lastVisionCamId;
		pInput->vision.updated = 1;

		lastVisionTimestamp = network.lastVisionPosTimestamp;
	}

	// KD
	pKD->kickerLevel = ctrlImpl.lastKdFeedback.kicker.cap;
	pKD->kickerChgCurrent = ctrlImpl.lastKdFeedback.kicker.chg;
	pKD->kickCounter = ctrlImpl.lastKdFeedback.kicker.kickCounter;

	pKD->dribblerSpeed = (int16_t)ctrlImpl.lastKdFeedback.dribbler.velocity;
	pKD->dribblerTemp = ctrlImpl.lastKdFeedback.dribbler.temp;
	pKD->dribblerOverheated = ctrlImpl.lastKdFeedback.dribbler.overheated;

	pKD->irLevel[0] = ctrlImpl.lastKdFeedback.barrier.level[0];
	pKD->irLevel[1] = ctrlImpl.lastKdFeedback.barrier.level[0];
	pKD->barrierInterrupted = ctrlImpl.lastKdFeedback.barrier.interrupted;

	pKD->header.timestamp = SysTimeUSec();
}

void CtrlImplApplyOutput(const CtrlOutput* pOutput)
{
	motors.ctrlMode = pOutput->ctrlMode;

	for(uint8_t i = 0; i < 4; i++)
	{
		motors.mot[i].set.velocity = pOutput->motorVel[i] * 1/(2*M_PI);
		motors.mot[i].set.voltage = pOutput->motorVoltage[i];
		motors.mot[i].set.torque = 0;
	}
}

void CtrlImplWriteLogMsg(void* pMsg)
{
	LogEntryHeader* pHeader = (LogEntryHeader*)pMsg;

	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_LOGFILE;
	iHeader.command = IC_LOGFILE_MESSAGE;

	IntercomSend(&iHeader, pHeader->length, pMsg);
}

void CtrlImplSendNetworkPacket(PacketHeader* pHeader, uint8_t* pData, uint16_t dataLength)
{
	NetworkSendPacket(pHeader, pData, dataLength);
}

void CtrlImplKickerFire(uint8_t mode, uint8_t device, float speed)
{
	KdFire fire = {mode, device, speed};
	IntercomHeader iHeader = { { INTERCOM_ADDRESS_KD, INTERCOM_SECTION_KD, IC_KD_FIRE } };

	IntercomReliableSend(&iHeader, sizeof(KdFire), &fire);
}

void CtrlImplBuzzerPlay(uint16_t seqId)
{
	switch(seqId)
	{
		case CTRL_BUZZ_UP20: BuzzerPlay(&buzzSeqUp20); break;
		case CTRL_BUZZ_BEEP_FAST: BuzzerPlay(&buzzSeqBeepFast); break;
		case CTRL_BUZZ_UP50: BuzzerPlay(&buzzSeqUp50); break;
		case CTRL_BUZZ_DOWN50: BuzzerPlay(&buzzSeqDown50); break;
		case CTRL_BUZZ_UPDOWN20: BuzzerPlay(&buzzSeqUpDown20); break;
	}
}

void CtrlImplCorrectOutXYW(float* pLocalVel)
{
	const int32_t numSupportPoints = sizeof(out2MotX)/sizeof(out2MotX[0]);

	float x = pLocalVel[0];
	float y = pLocalVel[1];

	float mod = 1;
	if(x < 0)
		mod = -1;

	float speed = sqrtf(x*x+y*y);

	if(speed < 0.001f)
		return;

	float dir = arm_atan2_f32(y, x*mod);

	float exactPos = (dir+PI/2)/PI*(numSupportPoints-1); // 0.0 - 1.0 => 0 - 129

	int32_t leftIndex = (int32_t)exactPos;
	int32_t rightIndex = (int32_t)(exactPos+1);

	float rightUsage = exactPos - leftIndex;
	float leftUsage = 1.0f - rightUsage;

	if(leftIndex < 0)
	{
		leftIndex = 0;
		rightIndex = 1;
	}

	if(rightIndex >= numSupportPoints)
	{
		rightIndex = numSupportPoints-1;
		leftIndex = numSupportPoints-2;
	}

	float corr[3];
	corr[0] = leftUsage*out2MotX[leftIndex] + rightUsage*out2MotX[rightIndex];
	corr[1] = leftUsage*out2MotY[leftIndex] + rightUsage*out2MotY[rightIndex];
	corr[2] = leftUsage*out2MotW[leftIndex] + rightUsage*out2MotW[rightIndex];

	pLocalVel[0] = speed*corr[0]*mod;
	pLocalVel[1] = speed*corr[1];
	pLocalVel[2] = pLocalVel[2] + speed*corr[2]*mod;
}

void CtrlImplCorrectInXYW(float* pLocalVel)
{
	const int32_t numSupportPoints = sizeof(out2MotX)/sizeof(out2MotX[0]);

	float x = pLocalVel[0];
	float y = pLocalVel[1];

	float mod = 1;
	if(x < 0)
		mod = -1;

	float speed = sqrtf(x*x+y*y);

	if(speed < 0.001f)
		return;

	float dir = arm_atan2_f32(y, x*mod);

	float exactPos = (dir+PI/2)/PI*(numSupportPoints-1); // 0.0 - 1.0 => 0 - 129

	int32_t leftIndex = (int32_t)exactPos;
	int32_t rightIndex = (int32_t)(exactPos+1);

	float rightUsage = exactPos - leftIndex;
	float leftUsage = 1.0f - rightUsage;

	if(leftIndex < 0)
	{
		leftIndex = 0;
		rightIndex = 1;
	}

	if(rightIndex >= numSupportPoints)
	{
		rightIndex = numSupportPoints-1;
		leftIndex = numSupportPoints-2;
	}

	float corr[3];
	corr[0] = leftUsage*mot2InX[leftIndex] + rightUsage*mot2InX[rightIndex];
	corr[1] = leftUsage*mot2InY[leftIndex] + rightUsage*mot2InY[rightIndex];
	corr[2] = leftUsage*out2MotW[leftIndex] + rightUsage*out2MotW[rightIndex];

	pLocalVel[0] = speed*corr[0]*mod;
	pLocalVel[1] = speed*corr[1];
	pLocalVel[2] = pLocalVel[2] - speed*corr[2]*mod;
}

static const float out2MotX[129] = {
0.00000000, 0.02286674, 0.04571869, 0.06854109, 0.09131917, 0.11446561, 0.13753832, 0.16052264,
0.18340399, 0.20682370, 0.23011144, 0.25325281, 0.27623353, 0.30057550, 0.32472852, 0.34867862,
0.37241192, 0.39889549, 0.42513495, 0.45111686, 0.47682795, 0.50686012, 0.53659539, 0.56602098,
0.59512431, 0.62750829, 0.65954543, 0.69122373, 0.72253145, 0.75218237, 0.78143984, 0.81029278,
0.83873038, 0.86341727, 0.88766791, 0.91147222, 0.93482042, 0.95503826, 0.97478123, 0.99404038,
1.01280705, 1.02917632, 1.04503668, 1.06038038, 1.07519997, 1.08804636, 1.10035471, 1.11211853,
1.12333166, 1.13290586, 1.14191806, 1.15036311, 1.15823620, 1.16472908, 1.17064142, 1.17596945,
1.18070975, 1.18443194, 1.18756064, 1.19009349, 1.19202851, 1.19317692, 1.19372460, 1.19367065,
1.19301451, 1.19166314, 1.18970960, 1.18715441, 1.18399850, 1.17997230, 1.17534827, 1.17012840,
1.16431504, 1.15717624, 1.14944971, 1.14113888, 1.13224750, 1.12156024, 1.11030102, 1.09847465,
1.08608628, 1.07186661, 1.05709625, 1.04178137, 1.02592843, 1.00872721, 0.99100187, 0.97275987,
0.95400896, 0.93499071, 0.91547997, 0.89548542, 0.87501599, 0.85505842, 0.83464474, 0.81378473,
0.79248847, 0.77192149, 0.75093916, 0.72955231, 0.70777201, 0.68621782, 0.66429304, 0.64200944,
0.61937899, 0.59622987, 0.57275850, 0.54897744, 0.52489949, 0.49990443, 0.47463855, 0.44911512,
0.42334759, 0.39710495, 0.37064552, 0.34398316, 0.31713185, 0.29060638, 0.26392026, 0.23708777,
0.21012331, 0.18376815, 0.15731000, 0.13076347, 0.10414321, 0.07815912, 0.05213071, 0.02607275,
0.00000000,  };

static const float out2MotY[129] = {
-1.07294922, -1.07340412, -1.07325684, -1.07250792, -1.07115828, -1.06851794, -1.06527977, -1.06144576,
-1.05701825, -1.05178365, -1.04596133, -1.03955469, -1.03256751, -1.02429923, -1.01545900, -1.00605161,
-0.99608222, -0.98396676, -0.97130062, -0.95808994, -0.94434122, -0.92754260, -0.91021987, -0.89238048,
-0.87403218, -0.85316381, -0.83180296, -0.80995828, -0.78763874, -0.76569255, -0.74329026, -0.72044164,
-0.69715677, -0.67568756, -0.65380301, -0.63151394, -0.60883141, -0.58774344, -0.56628488, -0.54446749,
-0.52230325, -0.50109248, -0.47955944, -0.45771673, -0.43557711, -0.41397496, -0.39210198, -0.36997147,
-0.34759685, -0.32536569, -0.30291776, -0.28026689, -0.25742707, -0.23451230, -0.21143688, -0.18821510,
-0.16486133, -0.14122956, -0.11749481, -0.09367167, -0.06977480, -0.04553011, -0.02124110, 0.00307746,
0.02741080, 0.05210187, 0.07677815, 0.10142488, 0.12602730, 0.15086850, 0.17563597, 0.20031506,
0.22489116, 0.24950126, 0.27397938, 0.29831114, 0.32248224, 0.34635958, 0.37004796, 0.39353342,
0.41680209, 0.43949497, 0.46194375, 0.48413499, 0.50605541, 0.52724202, 0.54813172, 0.56871175,
0.58896952, 0.60849946, 0.62768256, 0.64650683, 0.66496050, 0.68298596, 0.70061796, 0.71784544,
0.73465757, 0.75143046, 0.76776708, 0.78365739, 0.79909158, 0.81497766, 0.83038887, 0.84531626,
0.85975116, 0.87498961, 0.88971915, 0.90393203, 0.91762079, 0.93199043, 0.94582201, 0.95910907,
0.97184544, 0.98415108, 0.99589472, 1.00707121, 1.01767574, 1.02624465, 1.03423302, 1.04163708,
1.04845342, 1.05291430, 1.05678170, 1.06005324, 1.06272696, 1.06601354, 1.06869939, 1.07078360,
1.07226562,  };

static const float out2MotW[129] = {
-0.00365328, -0.00327173, -0.00210294, -0.00009425, 0.00277532, 0.00649057, 0.01099748, 0.01620242,
0.02197357, 0.02814442, 0.03451927, 0.04088052, 0.04699741, 0.05263572, 0.05756807, 0.06158417,
0.06450049, 0.06616898, 0.06648415, 0.06538826, 0.06287429, 0.05898655, 0.05381870, 0.04750955,
0.04023656, 0.03220754, 0.02365090, 0.01480496, 0.00590690, -0.00281810, -0.01116694, -0.01896745,
-0.02608536, -0.03242934, -0.03795348, -0.04265741, -0.04658393, -0.04981432, -0.05246151, -0.05466180,
-0.05656516, -0.05832513, -0.06008850, -0.06198566, -0.06412194, -0.06657060, -0.06936781, -0.07250990,
-0.07595311, -0.07961579, -0.08338300, -0.08711331, -0.09064721, -0.09381702, -0.09645743, -0.09841617,
-0.09956433, -0.09980556, -0.09908363, -0.09738805, -0.09475714, -0.09127857, -0.08708708, -0.08235959,
-0.07730781, -0.07216875, -0.06719356, -0.06263524, -0.05873594, -0.05571433, -0.05375392, -0.05299270,
-0.05351492, -0.05534521, -0.05844559, -0.06271538, -0.06799423, -0.07406794, -0.08067708, -0.08752775,
-0.09430416, -0.10068230, -0.10634418, -0.11099175, -0.11436010, -0.11622904, -0.11643274, -0.11486675,
-0.11149235, -0.10633782, -0.09949671, -0.09112324, -0.08142504, -0.07065366, -0.05909336, -0.04704871,
-0.03483161, -0.02274852, -0.01108822, -0.00011100, 0.00996046, 0.01894807, 0.02672289, 0.03320712,
0.03837360, 0.04224293, 0.04487839, 0.04637918, 0.04687233, 0.04650374, 0.04542904, 0.04380464,
0.04177959, 0.03948854, 0.03704632, 0.03454426, 0.03204840, 0.02959969, 0.02721590, 0.02489527,
0.02262132, 0.02036865, 0.01810918, 0.01581840, 0.01348116, 0.01109666, 0.00868209, 0.00627492,
0.00393336,  };

static const float mot2InX[129] = {
0.00108797, 0.02715690, 0.05318788, 0.07916313, 0.10505575, 0.13082927, 0.15643764, 0.18182585,
0.20693101, 0.23168399, 0.25601141, 0.27983798, 0.30308905, 0.32569320, 0.34758486, 0.36870666,
0.38901156, 0.40846455, 0.42704388, 0.44474177, 0.46156454, 0.47753221, 0.49267755, 0.50704466,
0.52068717, 0.53366607, 0.54604746, 0.55790011, 0.56929310, 0.58029361, 0.59096491, 0.60136467,
0.61154367, 0.62154485, 0.63140282, 0.64114377, 0.65078569, 0.66033899, 0.66980726, 0.67918828,
0.68847509, 0.69765707, 0.70672089, 0.71565149, 0.72443270, 0.73304780, 0.74147974, 0.74971130,
0.75772491, 0.76550244, 0.77302482, 0.78027168, 0.78722100, 0.79384884, 0.80012922, 0.80603417,
0.81153404, 0.81659797, 0.82119462, 0.82529311, 0.82886404, 0.83188065, 0.83432003, 0.83616422,
0.83740123, 0.83802580, 0.83803994, 0.83745314, 0.83628217, 0.83455055, 0.83228765, 0.82952739,
0.82630672, 0.82266382, 0.81863620, 0.81425872, 0.80956171, 0.80456927, 0.79929777, 0.79375477,
0.78793837, 0.78183698, 0.77542964, 0.76868679, 0.76157156, 0.75404141, 0.74605017, 0.73755024,
0.72849494, 0.71884089, 0.70855021, 0.69759250, 0.68594651, 0.67360127, 0.66055688, 0.64682460,
0.63242650, 0.61739458, 0.60176933, 0.58559800, 0.56893250, 0.55182711, 0.53433615, 0.51651172,
0.49840157, 0.48004731, 0.46148301, 0.44273423, 0.42381757, 0.40474075, 0.38550323, 0.36609719,
0.34650905, 0.32672114, 0.30671374, 0.28646703, 0.26596307, 0.24518764, 0.22413174, 0.20279275,
0.18117511, 0.15929060, 0.13715802, 0.11480241, 0.09225388, 0.06954597, 0.04671382, 0.02379209,
0.00081291,  };

static const float mot2InY[129] = {
-0.93197008, -0.93035305, -0.92868202, -0.92674410, -0.92439683, -0.92155245, -0.91815943, -0.91420377,
-0.90969344, -0.90465484, -0.89912708, -0.89314916, -0.88677188, -0.88003478, -0.87298474, -0.86565720,
-0.85807832, -0.85027882, -0.84226932, -0.83405673, -0.82564475, -0.81701958, -0.80817481, -0.79908417,
-0.78972418, -0.78006871, -0.77008119, -0.75972850, -0.74897534, -0.73778313, -0.72611886, -0.71394759,
-0.70123479, -0.68795149, -0.67407416, -0.65958069, -0.64444845, -0.62867164, -0.61223552, -0.59513940,
-0.57738905, -0.55898584, -0.53994341, -0.52028030, -0.50001774, -0.47917934, -0.45779484, -0.43589862,
-0.41352215, -0.39070514, -0.36748855, -0.34391172, -0.32001861, -0.29585270, -0.27145011, -0.24686000,
-0.22211669, -0.19726657, -0.17234115, -0.14737552, -0.12240679, -0.09746075, -0.07256588, -0.04774869,
-0.02302569, 0.00158343, 0.02606427, 0.05040524, 0.07459611, 0.09863579, 0.12251498, 0.14623507,
0.16980012, 0.19320509, 0.21645711, 0.23955858, 0.26251634, 0.28532818, 0.30800187, 0.33053735,
0.35293462, 0.37519205, 0.39730754, 0.41927331, 0.44107861, 0.46271071, 0.48415420, 0.50538724,
0.52639068, 0.54713735, 0.56759311, 0.58773073, 0.60751363, 0.62689878, 0.64585501, 0.66433710,
0.68229860, 0.69970695, 0.71651778, 0.73269174, 0.74819592, 0.76299598, 0.77706447, 0.79038153,
0.80293163, 0.81470315, 0.82570217, 0.83593379, 0.84541433, 0.85417522, 0.86225007, 0.86968243,
0.87652619, 0.88283882, 0.88868680, 0.89413601, 0.89925182, 0.90409827, 0.90872004, 0.91316314,
0.91743628, 0.92152910, 0.92538804, 0.92890604, 0.93192568, 0.93419739, 0.93538997, 0.93506155,
0.93263364,  };
