#include <main/ctrl.h>
#include <main/ctrl_tigga.h>
#include <main/skills.h>
#include "ch.h"

#include "hal/mco.h"
#include "hal/led.h"
#include "hal/buzzer.h"
#include "hal/exti.h"
#include "hal/usart1.h"
#include "hal/usart2.h"
#include "hal/usart6.h"

#include "util/boot.h"
#include "util/console.h"
#include "util/crc.h"
#include "util/flash_fs.h"
#include "util/init_hal.h"
#include "util/sys_time.h"
#include "util/network_print.h"
#include "util/terminal.h"
#include "util/intercom_reliable.h"
#include "util/prog_client.h"
#include "util/prog_store.h"

#include "cli.h"
#include "main/network.h"
#include "spi2.h"
#include "spi3.h"
#include "intercom1.h"
#include "hal/power.h"
#include "motors.h"
#include "ui.h"
#include "config_main.h"
#include "cpu_load.h"
#include "main/prog_source_wifi.h"

static THD_WORKING_AREA(_ccm_waUSART2, 512);
static THD_WORKING_AREA(_ccm_waSPI3, 512);
static THD_WORKING_AREA(_ccm_waSPI2, 512);
static THD_WORKING_AREA(_ccm_waUSART6, 512);
static THD_WORKING_AREA(_ccm_waIntercom, 512);
static THD_WORKING_AREA(_ccm_waTerminal, 4096);
#ifdef ENV_RUN
static THD_WORKING_AREA(_ccm_waCLI, 8192);
static THD_WORKING_AREA(_ccm_waNetwork, 4096);
static THD_WORKING_AREA(_ccm_waNetworkRel, 4096);
static THD_WORKING_AREA(_ccm_waUI, 512);
static THD_WORKING_AREA(_ccm_waMotors, 4096);
static THD_WORKING_AREA(_ccm_waConfig, 512);
#ifndef DEBUG0
static THD_WORKING_AREA(_ccm_waCTRL, 8192);
#endif
#endif

int main()
{
#ifdef ENV_BOOT
	if(!BootIsBootloaderSelected())
		BootJumpToApplication();

	BootSetBootloaderSelected(0);
#endif

	// Enable important clocks
	RCC->AHB1ENR |= RCC_AHB1ENR_DMA1EN | RCC_AHB1ENR_DMA2EN | RCC_AHB1ENR_GPIOAEN
			| RCC_AHB1ENR_GPIOBEN | RCC_AHB1ENR_GPIOCEN | RCC_AHB1ENR_GPIODEN;
	RCC->APB2ENR |= RCC_APB2ENR_SYSCFGEN;

	SystemClockInitData clkInit;
	clkInit.pll.M = 12;
	clkInit.pll.N = 336;
	clkInit.pll.P = 2;
	clkInit.pll.Q = 7;
	clkInit.HSEBypass = 1;
	clkInit.APB1Div = 4;	// = 42MHz
	clkInit.APB2Div = 4;	// = 42MHz
	clkInit.RTCDiv = 12;	// = 1MHz
	clkInit.flashLatency = 5;
	clkInit.sysTickFreq = CH_CFG_ST_FREQUENCY;
	SystemClockInit(&clkInit, 168000000UL);

	// Enable compensation cell to reduce noise on power lines with 50MHz/100MHz I/Os
	SYSCFG->CMPCR = SYSCFG_CMPCR_CMP_PD;

	// Enable usage fault, bus fault, and mem fault
	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk;

#ifdef ENV_BOOT
	MPU->CTRL = 0;
#else
	MPU->CTRL = MPU_CTRL_ENABLE_Msk | MPU_CTRL_PRIVDEFENA_Msk;
	// Region number: 0
	MPU->RNR = 0;
	// Base address: flash
	MPU->RBAR = 0x08000000;
	// Size: 1MB, disable for 384k-512k (FlashFS, sub-region 3), Read-Only
	MPU->RASR = MPU_RASR_ENABLE_Msk | (19 << 1) | (0b11111000 << 8) | MPU_RASR_C_Msk | (0x7 << 24);
#endif

//	SCB->CCR |= SCB_CCR_DIV_0_TRP_Msk;

	// set vector table offset
#ifdef ENV_BOOT
	SCB->VTOR = 0;
#else
	SCB->VTOR = BOOT_APP_OFFSET;
#endif

	chSysInit();

	// --- HAL ---
	LEDInit();		// init LEDs to signal RTOS startup failures
	PowerInit();
	PowerEnable5VSupply(1);
	FlashFSInit(0x08060000, 0x20000);	// 128k
	ConfigMainInit();
	BuzzerInit();
	ExtiInit();
	USART1Init();
	USART2Init();
	USART6Init();
	SPI2Init();
	SPI3Init();
	CRCInit();

	// --- High-Level Systems ---
	TerminalInit(&usart1.inQueue, &usart1.outQueue);
	Intercom1Init();
	IntercomReliableInit();
	ProgStoreInit();
	ProgClientInit();
	CLIInit();

#ifdef ENV_BOOT
//	ConsolePrint("\f--- Bootloader (MAIN) ---\r\n");
#else
	ConsolePrint("\f--- Main ---\r\n");
#endif

#ifdef ENV_RUN
	ProgSourceWifiInit();
	SkillsInit();
	NetworkInit();
	NetworkPrintInit(&NetworkSendPacketRaw, INTERCOM_ADDRESS_MAIN);
	MotorsInit();
	UIInit();
	CtrlInit();
	CtrlTiggaInitCfg();
	CtrlSetController(&ctrlTiggaInstance);
#endif

	// now start motor left/right and kicker processor
	MCOInit();

#ifdef ENV_RUN
	chThdCreateStatic(_ccm_waCLI, sizeof(_ccm_waCLI), NORMALPRIO-2, CLITask, 0);
#ifndef DEBUG0
	chThdCreateStatic(_ccm_waCTRL, sizeof(_ccm_waCTRL), NORMALPRIO+5, CtrlTask, 0);
#endif
	chThdCreateStatic(_ccm_waMotors, sizeof(_ccm_waMotors), NORMALPRIO-5, MotorsTask, 0);
	chThdCreateStatic(_ccm_waNetwork, sizeof(_ccm_waNetwork), NORMALPRIO+1, NetworkTask, 0);
	chThdCreateStatic(_ccm_waNetworkRel, sizeof(_ccm_waNetworkRel), NORMALPRIO-4, NetworkReliableTask, 0);
	chThdCreateStatic(_ccm_waConfig, sizeof(_ccm_waConfig), NORMALPRIO-8, ConfigTask, 0);
#endif
	chThdCreateStatic(_ccm_waUSART2, sizeof(_ccm_waUSART2), NORMALPRIO, USARTTask, &usart2);
	chThdCreateStatic(_ccm_waUSART6, sizeof(_ccm_waUSART6), NORMALPRIO, USARTTask, &usart6);
	chThdCreateStatic(_ccm_waSPI3, sizeof(_ccm_waSPI3), NORMALPRIO+10, SPI3Task, 0);
	chThdCreateStatic(_ccm_waIntercom, sizeof(_ccm_waIntercom), NORMALPRIO-1, IntercomTask, 0);
	chThdCreateStatic(_ccm_waTerminal, sizeof(_ccm_waTerminal), NORMALPRIO-2, TerminalTask, 0);
	chThdCreateStatic(_ccm_waSPI2, sizeof(_ccm_waSPI2), NORMALPRIO-3, SPI2Task, 0);

#ifdef ENV_RUN
	chThdCreateStatic(_ccm_waUI, sizeof(_ccm_waUI), NORMALPRIO-10, UITask, 0);
#endif

	CtrlSetFeatures((ctrl.misc.features & (~SYSTEM_MATCH_FEEDBACK_FEATURE_V2016)));

	if(IntercomGetReachableAddresses() & INTERCOM_ADDRESS_MEDIA)
		CtrlSetFeatures(ctrl.misc.features | SYSTEM_MATCH_FEEDBACK_FEATURE_EXT_BOARD);
	else
		CtrlSetFeatures(ctrl.misc.features & (~SYSTEM_MATCH_FEEDBACK_FEATURE_EXT_BOARD));

/*
	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);
*/

	LEDMustOff(LED_GREEN | LED_RED);

#ifndef ENV_RUN
	ProgStoreFlash();
#endif

	chThdSetPriority(HIGHPRIO-1);

	while(1)
	{
		chThdSleepMilliseconds(10);

#ifdef ENV_RUN
		uint32_t time = SysTimeUSec()%1000000;

		if(time < 200000)
			LEDMustOn(LED_GREEN);
		else
			LEDMustOff(LED_GREEN);
#else
		uint32_t time = SysTimeUSec()%500000;

		if(time < 100000)
			LEDMustOn(LED_RED);
		else
			LEDMustOff(LED_RED);
#endif

		CPULoadUpdateUsage();
	}

	return 0;
}
