/*
 * cli.c
 *
 *  Created on: 29.05.2010
 *      Author: AndreR
 */

#include "cli.h"

#include <main/ctrl_motor.h>
#include <main/skills.h>

#include "hal/led.h"
#include "hal/uart8.h"
#include "hal/buzzer.h"
#include "hal/tft.h"
#include "hal/motors.h"
#include "hal/kicker.h"
#include "hal/ext_flash.h"

#include "util/console.h"
#include "util/flash_fs.h"
#include "util/flash.h"
#include "util/network_print.h"
#include "util/terminal.h"
#include "util/boot.h"
#include "util/sys_time.h"
#include "util/log.h"
#include "util/crc.h"
#include "util/trajectory.h"
#include "util/trajectory_orient.h"
#include "util/traj_generator.h"
#include "util/arm_mat_util_f32.h"
#include "util/intercom_reliable.h"
#include "util/hermite_spline_cubic.h"
#include "util/ukf.h"
#include "util/angle_math.h"
#include "util/init_hal.h"
#include "util/prog_store.h"
#include "main/network.h"
#include "util/srukf.h"

#include "commands.h"
#include "intercom_constants.h"
#include "cpu_load.h"
#include "version.h"
#include "constants.h"
#include "spi4.h"
#include "spi6.h"
#include "power.h"
#include "gfx.h"
#include "fatfs/ff.h"
#include "ctrl_impl.h"
#include "util/log_file.h"
#include "main/ctrl_tigga.h"

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

CLIGlobal cli;

static void processCommand();

#define COLOR_NONE	0
#define COLOR_RED	1
#define COLOR_GREEN	2
#define COLOR_BLUE	3
#define COLOR_YELLOW 4

#define COLOR_MAIN	COLOR_NONE
#define COLOR_EXT	COLOR_BLUE

static void echoColor(uint8_t color)
{
	const char* pCol;

	switch(color)
	{
	case COLOR_RED:
		pCol = "\e[0;31m";
		break;
	case COLOR_GREEN:
		pCol = "\e[0;32m";
		break;
	case COLOR_BLUE:
		pCol = "\e[0;36m";
		break;
	case COLOR_YELLOW:
		pCol = "\e[0;33m";
		break;
	default:
		pCol = "\e[0m";
		break;
	}

	TerminalWrite((uint8_t*)pCol, strlen(pCol));
}

static void echoTargetColor(uint8_t addr)
{
	switch(addr)
	{
	case INTERCOM_ADDRESS_EXT:
		echoColor(COLOR_EXT);
		break;
	default:
		echoColor(COLOR_MAIN);
		break;
	}
}

static void consoleTxCallback(void* param)
{
	(void)param;

	while(console.txFifo.numPackets > 0)
	{
		echoTargetColor(INTERCOM_ADDRESS_MAIN);

		uint8_t* pData;
		uint32_t len32;
		int16_t result = FifoLinGet(&console.txFifo, &pData, &len32);
		if(result)
			break;

		TerminalWrite(pData, len32);

		FifoLinDelete(&console.txFifo);

		echoTargetColor(cli.currentTarget);
	}
}

// IRQ Level: Task
static void intercomCallback(const IntercomHeader* pHeader, uint16_t length, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	// called after new console intercom rx
	echoTargetColor(pHeader->command);
	TerminalWrite(pData, length);
	echoTargetColor(cli.currentTarget);
}

static void cmdCallback(char* pBuf, uint16_t length)
{
	ConsolePrint("\r\n");

	// target selection logic
	if(strcmp(pBuf, "target main") == 0)
	{
		cli.currentTarget = INTERCOM_ADDRESS_MAIN;
		echoTargetColor(cli.currentTarget);
		ConsolePrint("New target: MAIN\r\n");
		return;
	}

	if(strcmp(pBuf, "target ext") == 0)
	{
		if(uart8.extInstalled == 0)
		{
			ConsolePrint("EXT not installed\r\n");
			return;
		}

		cli.currentTarget = INTERCOM_ADDRESS_EXT;
		echoTargetColor(cli.currentTarget);
		ConsolePrint("New target: EXT\r\n");
		return;
	}

	// check current target
	if(cli.currentTarget == INTERCOM_ADDRESS_MAIN)
	{
		// process onboard
		uint8_t* pData;
		int16_t result = FifoLinReserve(&console.rxFifo, length, &pData);
		if(result)
			return;

		memcpy(pData, pBuf, length);

		FifoLinCommit(&console.rxFifo, length);
	}
	else
	{
		// forward to intercom
		IntercomHeader header;
		header.address = cli.currentTarget;
		header.command = 0;
		header.section = INTERCOM_SECTION_CONSOLE;

		IntercomReliableSend(&header, length, (uint8_t*)pBuf);
	}
}

static void escCallback(VT100Result* pResult)
{
	if(pResult->code == FUNCTION_KEY)
	{
		switch(pResult->param1)
		{
			case FK_F1:
			{
				cli.currentTarget = INTERCOM_ADDRESS_MAIN;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: MAIN\r\n");
			}
			break;
			case FK_F2:
			{
				if(uart8.extInstalled == 0)
				{
					ConsolePrint("EXT not installed\r\n");
					return;
				}

				cli.currentTarget = INTERCOM_ADDRESS_EXT;
				echoTargetColor(cli.currentTarget);
				ConsolePrint("New target: EXT\r\n");
			}
			break;
		}
	}
}

void CLIInit()
{
	cli.currentTarget = INTERCOM_ADDRESS_MAIN;

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = &consoleTxCallback;
	conData.pUserData = 0;

	ConsoleInit(conData);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &intercomCallback;
	client.section = INTERCOM_SECTION_CONSOLE;

	IntercomAddClient(client);

	terminal.cmdCb = &cmdCallback;
	terminal.escCb = &escCallback;
}

void CLITask(void* params)
{
	(void)params;

	chRegSetThreadName("CLI");

//	NetworkPrint("Firmware: %s - %s - %.7s\r\n", __DATE__, __TIME__,GIT_REF);

	while(1)
	{
		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

		if(cli.intercomFwd.inUse)
		{
			// forward to intercom
			IntercomHeader header;
			header.address = cli.intercomFwd.target;
			header.command = 0;
			header.section = INTERCOM_SECTION_CONSOLE;

			IntercomReliableSend(&header, cli.intercomFwd.bufSize, cli.intercomFwd.buf);

			cli.intercomFwd.inUse = 0;
		}

		chThdSleepMilliseconds(5);
	}
}

extern memory_heap_t guiHeap;

static void miscCommands()
{
	if(ConsoleCmpCmd("reset"))
	{
		BootReset();
	}

	if(ConsoleCmpCmd("fault"))
	{
		volatile uint32_t* pFault = (volatile uint32_t*)0x55555555;
		*pFault = 42;
	}

	if(ConsoleCmpCmd("bootloader"))
	{
		ConsolePrint("Selecting bootloader\r\n");

		BootSetBootloaderSelected(1);

//		IntercomHeader iHeader;
//		iHeader.section = INTERCOM_SECTION_BOOTLOADER;
//		iHeader.address = INTERCOM_ADDRESS_MAIN;
//		if(uart8.extInstalled)
//			iHeader.address |= INTERCOM_ADDRESS_EXT;
//		iHeader.command = BOOTLOADER_SELECT;
//
//		IntercomReliableSend(&iHeader, 0, 0);
	}

	if(ConsoleCmpCmd("shutdown"))
	{
		PowerShutdown();
	}

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 42);
	}

	float f = 21.0f;
	if(ConsoleScanCmd("float %f", &f) == 1)
	{
		ConsolePrint("Float: %f\r\n", f);
	}

	if(ConsoleCmpCmd("float"))
	{
		ConsolePrint("Float: %f\r\n", 42.0f);
	}

	if(ConsoleCmpCmd("cpu"))
	{
		CPULoadPrintUsage();
	}

	if(ConsoleCmpCmd("tasks"))
	{
		ConsolePrint("IRQ stack free: %u\r\n", chThdGetFreeStack(0));

		CPULoadPrintTasks();
	}

	if(ConsoleCmpCmd("irqs"))
	{
		ConsolePrint("Listing active IRQs...\r\n");

		for(int32_t i = -15; i < 0; i++)
		{
			ConsolePrint("%d: %u\r\n", i, NVIC_GetPriority(i));
			chThdSleepMilliseconds(10);
		}

		for(int32_t i = 0; i < 98; i++)
		{
			uint32_t enabled = NVIC->ISER[(((uint32_t)(int32_t)i) >> 5UL)] & ((uint32_t)(1UL << (((uint32_t)(int32_t)i) & 0x1FUL)));

			if(enabled)
			{
				ConsolePrint("%d: %u\r\n", i, NVIC_GetPriority(i));
				chThdSleepMilliseconds(10);
			}
		}
	}

	if(ConsoleCmpCmd("integ"))
	{
		chSysLock();
		uint8_t result = chSysIntegrityCheckI(CH_INTEGRITY_RLIST | CH_INTEGRITY_VTLIST | CH_INTEGRITY_REGISTRY | CH_INTEGRITY_PORT);
		chSysUnlock();

		if(result)
			ConsolePrint("Integrity check failed\r\n");
		else
			ConsolePrint("Integrity check OK\r\n");
	}

	if(ConsoleCmpCmd("heap"))
	{
		size_t free;
		size_t fragments = chHeapStatus(&guiHeap, &free);

		ConsolePrint("Fragments: %u, Free: %u, Core: %u\r\n", fragments, free, chCoreGetStatusX());
	}

	if(ConsoleCmpCmd("ttest"))
	{
//		thread_t* pThread = chRegNextThread(chRegNextThread(chRegFirstThread()));
//		thread_t* pThread = chRegNextThread(chRegFirstThread());
		thread_t* pThread = chRegFirstThread();
		uint32_t thdAddr = (uint32_t)pThread;

		uint8_t* pEnd = (uint8_t*)pThread->p_ctx.r13;
		uint8_t* pStack = (uint8_t*)(pThread+1);
		while(pStack < pEnd && *pStack == CH_DBG_STACK_FILL_VALUE)
			pStack++;

		uint32_t stackFree = (uint32_t)pStack - (uint32_t)pThread;

		uint32_t endThd = ((uint32_t)pThread->p_ctx.r13)+sizeof(struct port_intctx);
		uint32_t thdSize = endThd-(uint32_t)pThread;

		ConsolePrint("%s\r\n", pThread->p_name);
		ConsolePrint("pThd: 0x%08X\r\n", thdAddr);
		ConsolePrint("Stack start: 0x%08X\r\n", thdAddr+sizeof(thread_t));
		ConsolePrint("R13: 0x%08X\r\n", (uint32_t)pThread->p_ctx.r13);
		ConsolePrint("Thd: %u, IntCtx: %u, ExtCtx: %u\r\n", sizeof(thread_t), sizeof(struct port_intctx), sizeof(struct port_extctx)); // +PORT_INT_REQUIRED_STACK
		ConsolePrint("waSize: %u\r\n", thdSize);
		ConsolePrint("free: %u\r\n", stackFree);
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());

		struct tm* nowtm;
		time_t now = SysTimeUnix();
		nowtm = localtime(&now);

		char tmbuf[64];
		strftime(tmbuf, sizeof tmbuf, "%Y-%m-%d %H:%M:%S", nowtm);
		ConsolePrint("%s\r\n", tmbuf);
	}

	uint16_t u16;
	if(ConsoleScanCmd("log %hu", &u16) == 1)
	{
		LogPrintAllRecent(u16);
	}

	if(ConsoleCmpCmd("spi6"))
	{
		ConsolePrint("V_bat: %fV\r\n", power.vBat);
		ConsolePrint("I_cur: %fA\r\n", power.iCur);
		ConsolePrint("V_ir:  %fV\r\n", spi6.vIr);
		ConsolePrint("V_irOn:  %fV\r\n", spi6.vIrOn);
		ConsolePrint("V_irOff:  %fV\r\n", spi6.vIrOff);
		ConsolePrint("Acc: %.4f, %.4f, %.4f\r\n", spi6.imu.acc[0], spi6.imu.acc[1], spi6.imu.acc[2]);
		ConsolePrint("Gyr: %.4f, %.4f, %.4f\r\n", spi6.imu.gyr[0], spi6.imu.gyr[1], spi6.imu.gyr[2]);
		ConsolePrint("Temp: %.2f\r\n", spi6.imu.temp);
		ConsolePrint("t_pwr:  %uus\r\n", spi6.power.sampleTime);
		ConsolePrint("t_kick: %uus\r\n", spi6.kick.sampleTime);
		ConsolePrint("t_imu:  %uus\r\n", spi6.imu.sampleTime);
	}

	if(ConsoleCmpCmd("spi4"))
	{
		ConsolePrint("-- Touch --\r\n");
		ConsolePrint("Pressed: %hu\r\nX: %hu\r\nY: %hu\r\n", (uint16_t)spi4.touch.pressed,
				spi4.touch.x, spi4.touch.y);
		ConsolePrint("t_touch: %uus\r\n", spi4.touch.sampleTime);
		chThdSleepMilliseconds(20);
		ConsolePrint("-- MSL --\r\n");
		ConsolePrint("Functional: %hu\r\n", (uint16_t)spi4.mouse[0].functional);
		ConsolePrint("Motion: 0x%02hX\r\n", (uint16_t)spi4.mouse[0].motion.motion);
		ConsolePrint("dX: %hd\r\ndY: %hd\r\n", spi4.mouse[0].motion.deltaX, spi4.mouse[0].motion.deltaY);
		ConsolePrint("SQUAL: %hu\r\n", (uint16_t)spi4.mouse[0].motion.squal);
		ConsolePrint("pixelSum: %hu\r\n", (uint16_t)spi4.mouse[0].motion.pixelSum);
		ConsolePrint("maxPixel: %hu\r\n", (uint16_t)spi4.mouse[0].motion.maxPixel);
		ConsolePrint("minPixel: %hu\r\n", (uint16_t)spi4.mouse[0].motion.minPixel);
		ConsolePrint("Shutter: %hu\r\nFrame Period: %hu\r\n", spi4.mouse[0].motion.shutter, spi4.mouse[0].motion.framePeriod);
		chThdSleepMilliseconds(20);
		ConsolePrint("-- MSR --\r\n");
		ConsolePrint("Functional: %hu\r\n", (uint16_t)spi4.mouse[1].functional);
		ConsolePrint("Motion: 0x%02hX\r\n", (uint16_t)spi4.mouse[1].motion.motion);
		ConsolePrint("dX: %hd\r\ndY: %hd\r\n", spi4.mouse[1].motion.deltaX, spi4.mouse[1].motion.deltaY);
		ConsolePrint("SQUAL: %hu\r\n", (uint16_t)spi4.mouse[1].motion.squal);
		ConsolePrint("pixelSum: %hu\r\n", (uint16_t)spi4.mouse[1].motion.pixelSum);
		ConsolePrint("maxPixel: %hu\r\n", (uint16_t)spi4.mouse[1].motion.maxPixel);
		ConsolePrint("minPixel: %hu\r\n", (uint16_t)spi4.mouse[1].motion.minPixel);
		ConsolePrint("Shutter: %hu\r\nFrame Period: %hu\r\n", spi4.mouse[1].motion.shutter, spi4.mouse[1].motion.framePeriod);
		ConsolePrint("t_optical: %uus\r\n", spi4.mouseSampleTime);
	}

#ifndef ENV_BOOT
	if(ConsoleScanCmd("rot %hu", &u16) == 1)
	{
		switch(u16)
		{
		case 0: gdispSetOrientation(GDISP_ROTATE_0); break;
		case 1: gdispSetOrientation(GDISP_ROTATE_90); break;
		case 2: gdispSetOrientation(GDISP_ROTATE_180); break;
		case 3: gdispSetOrientation(GDISP_ROTATE_270); break;
		}

		gdispClear(Red);
		gdispFillArea(0, 0, 100, 200, Blue);
		gdispDrawString(10, 10, "Test", gdispOpenFont("fixed_7x14"), White);
	}
#endif

	if(ConsoleCmpCmd("sd"))
	{
		ConsolePrint("STA: 0x%08X\r\n", SDMMC1->STA);
		ConsolePrint("FIFOCNT: %u\r\n", SDMMC1->FIFOCNT);
	}

	if(ConsoleCmpCmd("qspi"))
	{
		ExtFlashPrintPerformance();

		ExtFlashChipErase();

		ConsolePrint("Writing...\r\n");

		for(uint32_t i = 0; i < 1*1024*1024; i += 4)
		{
			ExtFlashWrite(i, &i, 4);
		}

		ConsolePrint("Reading...\r\n");

		for(uint32_t i = 0; i < 1*1024*1024; i += 4)
		{
			uint32_t data;

			ExtFlashRead(i, &data, 4);

			uint32_t comp = i;

			if(data != comp)
			{
				ConsolePrint("ExtFlash mismatch at %u: 0x%08X, 0x%08X\r\n", i, i, data);
//				break;
			}
		}

		ConsolePrint("done\r\n");
	}

	if(ConsoleCmpCmd("flash list"))
	{
		FlashFSListFiles();
	}

	if(ConsoleCmpCmd("flash erase"))
	{
		ConsolePrint("Erasing flash\r\n");

		FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);
	}

	if(ConsoleCmpCmd("flash compact"))
	{
		FlashFSCompact();

		ConsolePrint("Flash compacted\r\n");
	}

	if(ConsoleCmpCmd("flash hex"))
	{
		for(uint32_t i = 0; i < flashFS.bytesDel+flashFS.bytesUsed+48; i += 8)
		{
			uint8_t* d = (uint8_t*)(flashFS.flashAddr+i);
			ConsolePrint("0x%04X: %02X %02X %02X %02X %02X %02X %02X %02X\r\n",
					i, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);

			chThdSleepMilliseconds(5);
		}
	}

	if(ConsoleCmpCmd("flash test"))
	{
		uint32_t buf[16];
		memset(buf, 0x42, 16*sizeof(uint32_t));
		int16_t result = FlashProgram(flashFS.flashAddr, buf, 8);

		ConsolePrint("FlashProgram: %hd\r\n", result);
	}

	if(ConsoleCmpCmd("prog wifi"))
	{
		ProgStoreStart(PROG_STORE_SOURCE_WIFI);
	}

	if(ConsoleCmpCmd("prog usb"))
	{
		ProgStoreStart(PROG_STORE_SOURCE_USB);
	}

	if(ConsoleScanCmd("nrf read %hu", &u16) == 1)
	{
		uint8_t reg = 0x42;

		int16_t result = NRF24ExReadRegister(&network.nrf24, u16, &reg);

		ConsolePrint("Result: %hd\r\n", result);
		ConsolePrint("Status: 0x%02hX\r\n", (uint16_t)network.nrf24.lastStatus);
		ConsolePrint("0x%02hX: 0x%02hX\r\n", u16, (uint16_t)reg);
	}

	if(ConsoleCmpCmd("dwt"))
	{
		ConsolePrint("DWT: %u\r\n", *DWT_CYCCNT);
	}

	if(ConsoleCmpCmd("configs"))
	{
		for(uint16_t i = 0; i < config.filesUsed; i++)
		{
			ConsolePrint("%hu: %s\r\n", i, config.files[i].pDesc->pName);
		}
	}

	if(ConsoleCmpCmd("qr"))
	{
		ConsolePrint("QR Decomposition Test\r\n");

//		arm_matrix_instance_f32 A = {3, 3, (float[9]){
//			12, -51, 4, 6, 167, -68, -4, 24, -41
//		}};
//		arm_matrix_instance_f32 Q = {3, 3, (float[9]){}};
//		arm_matrix_instance_f32 R = {3, 3, (float[9]){}};
//		arm_matrix_instance_f32 H = {3, 3, (float[9]){}};
//		arm_matrix_instance_f32 tmp = {3, 3, (float[9]){}};

		arm_matrix_instance_f32 A = {4, 3, (float[12]){
			12, -51, 4, 6, 167, -68, -4, 24, -41, 33, -54, 38
		}};
//		arm_matrix_instance_f32 A = {4, 3, (float[12]){
//			12, 0, 4, 6, 0, -68, -4, 0, -41, 33, 0, 38
//		}};
		arm_matrix_instance_f32 Q = {4, 4, (float[16]){}};
		arm_matrix_instance_f32 R = {4, 3, (float[12]){}};
		arm_matrix_instance_f32 H = {4, 4, (float[16]){}};
		arm_matrix_instance_f32 tmp = {4, 4, (float[16]){}};

		chSysLock();
		uint32_t start = *DWT_CYCCNT;

		arm_mat_qr(&A, 0, &R, &H, &tmp);
		uint32_t end = *DWT_CYCCNT;
		chSysUnlock();

		ConsolePrint("A:\r\n");
		arm_mat_print(&A);
		ConsolePrint("Q:\r\n");
		arm_mat_print(&Q);
		ConsolePrint("R:\r\n");
		arm_mat_print(&R);
		ConsolePrint("Cycles: %u\r\n", end-start);
	}

	if(ConsoleCmpCmd("cholup"))
	{
		ConsolePrint("Cholesky rank 1 update test\r\n");

		arm_matrix_instance_f32 L = {3, 3, (float[9]){
			-14, -21, 14, 0, -175, 70, 0, 0, 35
		}};
		float x[3] = {1, 2, 3};

		ConsolePrint("L:\r\n");
		arm_mat_print(&L);

		chSysLock();
		uint32_t start = *DWT_CYCCNT;
		arm_mat_cholupdate(&L, x);
//		arm_mat_choldowndate(&L, x);
		uint32_t end = *DWT_CYCCNT;
		chSysUnlock();

		ConsolePrint("L:\r\n");
		arm_mat_print(&L);

		ConsolePrint("Cycles: %u\r\n", end-start);
	}

	float f32[3];
	if(ConsoleScanCmd("out %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("Correcting out vel: %f, %f, %f\r\n", f32[0], f32[1], f32[2]);

		CtrlImplCorrectOutXYW(f32);

		ConsolePrint("Result: %f, %f, %f\r\n", f32[0], f32[1], f32[2]);
	}

	if(ConsoleScanCmd("in %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("Correcting in vel: %f, %f, %f\r\n", f32[0], f32[1], f32[2]);

		CtrlImplCorrectInXYW(f32);

		ConsolePrint("Result: %f, %f, %f\r\n", f32[0], f32[1], f32[2]);
	}

	if(ConsoleScanCmd("tigga D %f", f32) == 1)
	{
		ctrlTigga.trajGen.dampingXY = *f32;
	}

	if(ConsoleScanCmd("tigga t %f", f32) == 1)
	{
		ctrlTigga.trajGen.tShiftXY = *f32;
	}

	if(ConsoleCmpCmd("jerk1"))
	{
		TrajGenerator gen;

		TrajGeneratorInit(&gen);

		float filtPos[3] = {0, 0, 0};
		float filtVel[3] = {0, 0, 0};

		float posTrg[3] = {2, 0, 0};

		TrajGeneratorSetState(&gen, filtPos, filtVel);

		chSysLock();
		uint32_t start = *DWT_CYCCNT;

		TrajGeneratorCreateGlobalPos(&gen, posTrg);

//		TrajGeneratorStep(&gen, 0.002f);

		uint32_t end = *DWT_CYCCNT;
		chSysUnlock();

		ConsolePrint("Cycles: %u\r\n", end-start);

//		float p[3];
//		float v[3];
//		float a[3];
//		float j[3];
//
//		for(float t = 0; t < 2.0f; t += 0.1)
//		{
//			TrajGeneratorValuesAtTime(&gen, t, p, v, a, j);
//
//			ConsolePrint("%.2f\t%.3f,%.3f\t%.3f,%.3f\t%.3f,%.3f\t%.3f,%.3f\r\n",
//					t, p[0], p[1], v[0], v[1], a[0], a[1], j[0], j[1]);
//
//			chThdSleepMilliseconds(10);
//		}
//
//		ConsolePrint("---\r\n");

		const float* p = gen.trajGlobalPos;
		const float* v = gen.trajGlobalVel;
		const float* a = gen.trajGlobalAcc;
		const float* j = gen.trajGlobalJerk;

		for(float t = 0; t < 2.0f; t += 0.00125f)
		{
			TrajGeneratorCreateGlobalPos(&gen, posTrg);
			TrajGeneratorStep(&gen, 0.00125f);

			ConsolePrint("%.5f\t%.5f,%.5f\t%.5f,%.5f\t%.5f,%.5f\t%.5f,%.5f\r\n",
					t, p[0], p[1], v[0], v[1], a[0], a[1], j[0], j[1]);

			chThdSleepMilliseconds(10);
		}

//		for(float t = 0; t < 2.0f; t += 0.1)
//		{
//			TrajGeneratorValuesAtTime(&gen, t, p, v, a, j);
//
//			ConsolePrint("%.2f\t%.3f\t%.3f\t%.3f\t%.3f\r\n",
//					t, p[2], v[2], a[2], j[2]);
//
//			chThdSleepMilliseconds(10);
//		}
	}

	if(ConsoleCmpCmd("srukf"))
	{
		ConsolePrint("x:\r\n");
		arm_mat_print(&ctrlTigga.ukf.x);

		ConsolePrint("Sx:\r\n");
		arm_mat_print(&ctrlTigga.ukf.Sx);

		ConsolePrint("Vision: %f, %f, %f\r\n",
				ctrl.input.vision.pos[0], ctrl.input.vision.pos[1], ctrl.input.vision.pos[2]);

		ConsolePrint("States:\r\n");
		for(uint8_t i = 0; i < 8; i++)
			ConsolePrint("%u\r\n", ctrlTigga.ukf.pAngularState[i]);

		ConsolePrint("Meas:\r\n");
		for(uint8_t i = 0; i < 6; i++)
			ConsolePrint("%u\r\n", ctrlTigga.ukf.pAngularMeas[i]);

		ConsolePrint("Update:\r\n");
		for(uint8_t i = 0; i < 6; i++)
			ConsolePrint("%u\r\n", ctrlTigga.ukf.pUpdate[i]);

		return;
		float meas[6] = {1, 0, 0, 0, 0, 0};

		chSysLock();
		uint32_t start = *DWT_CYCCNT;

		SRUKFPredict(&ctrlTigga.ukf);

		memcpy(ctrlTigga.ukf.y.pData, meas, sizeof(float)*6);

		SRUKFUpdate(&ctrlTigga.ukf);

		uint32_t end = *DWT_CYCCNT;
		chSysUnlock();

//		arm_mat_print(&kf.x);

		ConsolePrint("Cycles: %u\r\n", end-start);
	}

	if(ConsoleCmpCmd("watchdog"))
	{
		ConsolePrint("Z: %f, W: %f\r\n", ctrl.accZ, ctrl.gyrW);
	}
}

static void wifiCommands()
{
	uint16_t u16;

	if(ConsoleCmpCmd("wifi stats"))
	{
		ConsolePrint("Avg: %u\r\n", network.avgRxPeriod);

		NRF24QueueStats* pStat = &network.queue.stats;

		ConsolePrint(" %2hu        toBS     fromBS     rxLoss     txLoss\r\n", (uint16_t)network.queue.id);
		ConsolePrint("LL P %10u %10u %10u %10u\r\n", pStat->nrfIO.txPackets, pStat->nrfIO.rxPackets,
				pStat->nrfFeedback.rxPacketsLost, pStat->nrfFeedback.txPacketsMaxRT);
		ConsolePrint("   B %10u %10u\r\n", pStat->nrfIO.txBytes, pStat->nrfIO.rxBytes);
		ConsolePrint("HL P %10u %10u %10u %10u\r\n", pStat->queueIO.txPackets, pStat->queueIO.rxPackets,
				pStat->queueFeedback.rxPacketsLost, pStat->queueFeedback.txPacketsLost);
		ConsolePrint("   B %10u %10u\r\n", pStat->queueIO.txBytes, pStat->queueIO.rxBytes);
	}

	if(ConsoleCmpCmd("color b"))
	{
		if(network.config.botId < 12)
		{
			network.config.botId += 12;
			NetworkSaveConfig();
			NetworkPrint("Changed BotID to %hu", u16);
		}
	}

	if(ConsoleCmpCmd("color y"))
	{
		if(network.config.botId > 11)
		{
			network.config.botId -= 12;
			NetworkSaveConfig();
			NetworkPrint("Changed BotID to %hu", u16);
		}
	}

	if(ConsoleScanCmd("set botid %hu", &u16) == 1)
	{
		network.config.botId = u16;
		NetworkSaveConfig();

		NetworkPrint("Changed BotID to %hu", u16);
		ConsolePrint("Changed BotID to %hu\r\n", u16);
	}

	if(ConsoleScanCmd("write botid %hu", &u16) == 1)
	{
		network.config.botId = u16;
		NetworkSaveConfig();

		NetworkPrint("Changed BotID to %hu", u16);
		ConsolePrint("Changed BotID to %hu\r\n", u16);
	}

	if(ConsoleScanCmd("set channel %hu", &u16) == 1)
	{
		network.config.channel = u16;
		NetworkSaveConfig();

		ConsolePrint("Changed channel to %hu\r\n", u16);
	}

	uint16_t speed;
	if(ConsoleScanCmd("set speed %hu", &speed) == 1)
	{
		network.config.speed = speed;
		NetworkSaveConfig();

		if(network.config.speed == 0)
			ConsolePrint("Speed: 250k\r\n");
		if(network.config.speed == 1)
			ConsolePrint("Speed: 1M\r\n");
		if(network.config.speed == 2)
			ConsolePrint("Speed: 2M\r\n");
	}

	if(ConsoleScanCmd("set features %hX", &speed) == 1)
	{
		CtrlSetFeatures(speed);

		ConsolePrint("Features: 0x%04hX\r\n", ctrl.misc.features);
	}

	if(ConsoleCmpCmd("config write"))
	{
		NetworkSaveConfig();
		ConsolePrint("Config saved to FLASH\r\n");
	}

	if(ConsoleCmpCmd("config"))
	{
		ConsolePrint("ID: %hu\r\n", network.config.botId);
		ConsolePrint("Channel: %hu\r\n", network.config.channel);
		if(network.config.speed == 0)
			ConsolePrint("Speed: 250k\r\n");
		if(network.config.speed == 1)
			ConsolePrint("Speed: 1M\r\n");
		if(network.config.speed == 2)
			ConsolePrint("Speed: 2M\r\n");

		ConsolePrint("HWID: %hu\r\n", (uint16_t)ctrl.misc.hardwareId);
		ConsolePrint("Features: 0x%04hX\r\n", ctrl.misc.features);
	}

	if(ConsoleScanCmd("set hwid %hu", &u16) == 1)
	{
		CtrlSetHardwareId(u16);

		ConsolePrint("New hardware ID: %hu\r\n", u16);
	}
}

static void kickerCommands()
{
	uint16_t u16;
	uint32_t u32;
	float f32;

	if(ConsoleScanCmd("vmax %hu", &u16) == 1)
	{
		ConsolePrint("vMax: %hu\r\n", u16);

		KickerSetMaxVoltage(u16);
	}

	if(ConsoleScanCmd("chg %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Enable charging\r\n");
			KickerEnableCharge(1);
		}
		else
		{
			ConsolePrint("Disable charging\r\n");
			KickerEnableCharge(0);
		}
	}

	if(ConsoleScanCmd("autochg %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Enable autocharging\r\n");
			KickerEnableAutoRecharge(1);
		}
		else
		{
			ConsolePrint("Disable autocharging\r\n");
			KickerEnableAutoRecharge(0);
		}
	}

	if(ConsoleCmpCmd("autodischg"))
	{
		ConsolePrint("Enabling auto discharging\r\n");
		KickerAutoDischarge();
	}

	if(ConsoleScanCmd("kstress %f", &f32) == 1)
	{
		ConsolePrint("Setting auto-fire stress test: %fms\r\n", f32);

		kicker.stress = f32*0.001f;
	}

	if(ConsoleScanCmd("dis1 %f", &f32) == 1)
	{
		ConsolePrint("Fire1: %fms\r\n", f32);

		KickerFire(f32*0.001f, KICKER_DEVICE_STRAIGHT);
	}

	if(ConsoleScanCmd("arm1 %f", &f32) == 1)
	{
		ConsolePrint("Arm1: %fms\r\n", f32);

		KickerArm(f32*0.001f, KICKER_DEVICE_STRAIGHT);
	}

	if(ConsoleScanCmd("dis2 %f", &f32) == 1)
	{
		ConsolePrint("Fire2: %fms\r\n", f32);

		KickerFire(f32*0.001f, KICKER_DEVICE_CHIP);
	}

	if(ConsoleScanCmd("arm2 %f", &f32) == 1)
	{
		ConsolePrint("Arm2: %fms\r\n", f32);

		KickerArm(f32*0.001f, KICKER_DEVICE_CHIP);
	}

	if(ConsoleCmpCmd("kstat"))
	{
		ConsolePrint("Installed: %hu\r\n", (uint16_t)kicker.installed);
		ConsolePrint("Cap: %.3fV (%.1fV)\r\n", kicker.vCap, kicker.config.maxVoltage);
		ConsolePrint("Chg: %.3fA\r\n", kicker.iChg);
		ConsolePrint("IR: %.3f / %.3f\r\n", kicker.vIrOff, kicker.vIrOn);
		ConsolePrint("State: %hu, Auto: %hu\r\n", (uint16_t)kicker.state, (uint16_t)kicker.autoCharge);
	}

	if(ConsoleScanCmd("cooldown %u", &u32) == 1)
	{
		KickerSetCooldown(u32);

		ConsolePrint("Kicker Cooldown set to %ums\r\n", kicker.config.cooldown);
	}
}

static void motorsCommands()
{
	uint16_t u16[4];
	float f32[4];

	if(ConsoleCmpCmd("ctrl off"))
	{
		CtrlSetController(0);

		ConsolePrint("Controller disabled\r\n");
	}

	if(ConsoleCmpCmd("ctrl motor"))
	{
		CtrlSetController(&ctrlMotorInstance);

		ConsolePrint("Controller: Motors\r\n");
	}

	if(ConsoleCmpCmd("skill off"))
	{
		uint8_t data[SYSTEM_MATCH_CTRL_USER_DATA_SIZE];

		SkillsSetInput(0, data);
	}

	if(ConsoleScanCmd("skill override %hu", u16) == 1)
	{
		if(u16[0])
		{
			ConsolePrint("Skill override enabled\r\n");
			NetworkPrint("Skill override enabled\n");
			network.skillOverride = 1;
		}
		else
		{
			ConsolePrint("Skill override disabled\r\n");
			NetworkPrint("Skill override disabled\n");
			network.skillOverride = 0;
		}
	}

	if(ConsoleScanCmd("move %f %f %f", f32, f32+1, f32+2) == 3)
	{
//		CtrlMotorSetVelTarget(f32[0], f32[1], f32[2]);
//		CtrlTiggaSetVelTarget(f32[0], f32[1], f32[2]);

		uint8_t data[SYSTEM_MATCH_CTRL_USER_DATA_SIZE];
		int16_t* p16Data = (int16_t*)data;
		p16Data[0] = (int16_t)(f32[0]*1000.0f);
		p16Data[1] = (int16_t)(f32[1]*1000.0f);
		p16Data[2] = (int16_t)(f32[2]*1000.0f);

		SkillsSetInput(1, data);
	}

	if(ConsoleScanCmd("goto %f %f %f", f32, f32+1, f32+2) == 3)
	{
		uint8_t data[SYSTEM_MATCH_CTRL_USER_DATA_SIZE];
		int16_t* p16Data = (int16_t*)data;
		p16Data[0] = (int16_t)(f32[0]*1000.0f);
		p16Data[1] = (int16_t)(f32[1]*1000.0f);
		p16Data[2] = (int16_t)(f32[2]*1000.0f);
		p16Data[3] = 0;

		SkillsSetInput(3, data);
	}

	if(ConsoleScanCmd("mot en %hu", u16) == 1)
	{
		ConsolePrint("Motors: %hu\r\n", u16[0]);

		MotorsDriveEnable(u16[0]);
	}

	if(ConsoleScanCmd("mot mode %hu", u16) == 1)
	{
		ConsolePrint("Motor mode: %hu\r\n", u16[0]);

		motors.driveMode = u16[0];
	}

	if(ConsoleCmpCmd("mot fault"))
	{
		MotorsDrivePrintFault();
	}

	if(ConsoleScanCmd("mot ccr %hu %hu %hu %hu", u16, u16+1, u16+2, u16+3) == 4)
	{
		ConsolePrint("%hu\r\n%hu\r\n%hu\r\n%hu\r\n", u16[0], u16[1], u16[2], u16[3]);

		MotorsDriveSetRaw(u16);
	}

	if(ConsoleScanCmd("mot vol %hu %f", u16, f32) == 2)
	{
		ConsolePrint("Motor %hu: %.3fV\r\n", u16[0], f32[0]);
		motors.drive[u16[0]].setVoltage = f32[0];
	}

	if(ConsoleScanCmd("mot vel %hu %f", u16, f32) == 2)
	{
		ConsolePrint("Vel %hu: %.3frad/s\r\n", u16[0], f32[0]);

		motors.drive[u16[0]].setVelocity = f32[0];
	}

	if(ConsoleScanCmd("mot pid %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("P: %f, I: %f, D: %f\r\n", f32[0], f32[1], f32[2]);

		for(uint8_t i = 0; i < 4; i++)
			PIDSetParams(&motors.drive[i].pid, f32[0], f32[1], f32[2]);
	}

	if(ConsoleCmpCmd("enc"))
	{
		MotorsDrivePrintEncoders();
	}

	if(ConsoleCmpCmd("abg"))
	{
		for(uint16_t i = 0; i < 4; i++)
		{
			ConsolePrint("%hu: %.6f\r\n", i, motors.drive[i].abg.x[1]);
		}
	}

	if(ConsoleCmpCmd("abgt"))
	{
		ABGFilter abg;
		ABGFilterInit(&abg, 1000, 50, ABG_FILTER_CRITICALLY_DAMPED);

		ConsolePrint("x: %f %f %f\r\n", abg.x[0], abg.x[1], abg.x[2]);

		for(uint16_t i = 0; i < 50; i++)
		{
			ABGFilterUpdate(&abg, 1.0f);
			ConsolePrint("x: %f %f %f\r\n", abg.x[0], abg.x[1], abg.x[2]);
			chThdSleepMilliseconds(1);
		}
	}

	if(ConsoleScanCmd("drib en %hu", u16) == 1)
	{
		ConsolePrint("Dribblers: %hu\r\n", u16[0]);

		MotorsDribblerEnable(u16[0]);
	}

	if(ConsoleScanCmd("drib ccr %hu", u16) == 1)
	{
		ConsolePrint("%hu\r\n", u16[0]);

		MotorsDribblerSetRaw(u16[0]);
	}

	if(ConsoleScanCmd("drib vol %f", f32) == 1)
	{
		ConsolePrint("Dribbler voltage: %f\r\n", f32[0]);

		MotorsDribblerSetVoltage(f32[0]);
	}

	if(ConsoleCmpCmd("drib stat"))
	{
		ConsolePrint("LastUpdate: %u\r\nVelocity: %f\r\n", motors.dribbler.lastUpdateTime,
				motors.dribbler.velocity*MOTORS_DRIBBLER_MOTOR_RADS_TO_BAR_RPM);
	}

	if(ConsoleScanCmd("drib pid %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("P: %f, I: %f, D: %f\r\n", f32[0], f32[1], f32[2]);

		PIDSetParams(&motors.dribbler.pid, f32[0], f32[1], f32[2]);
	}

	if(ConsoleScanCmd("drib vel %f", f32) == 1)
	{
		ConsolePrint("Dribbler speed: %fRPM\r\n", f32[0]);

		MotorsDribblerSetVelocity(f32[0]);
//		motors.dribbler.setVelocity = f32[0];
	}

	if(ConsoleScanCmd("drib ema %f", f32) == 1)
	{
		ConsolePrint("Dribbler EMA Alpha: %f\r\n", f32[0]);

		motors.dribbler.ema.alpha = f32[0];
	}

	if(ConsoleCmpCmd("mm"))
	{
		ConsolePrint("Mode: %hu\r\n", (uint16_t)motors.driveMode);
		ConsolePrint("Motors On: %hu\r\n", (uint16_t)ctrl.motorsOn);
	}

	if(ConsoleScanCmd("motors %hu", u16) == 1)
	{
		ConsolePrint("Motors: %hu\r\n", u16[0]);

		CtrlSetMotorsOn(u16[0]);
	}
}

static void fsCommands()
{
	FRESULT fresult;
	static const uint16_t bufLen = 128;
	static uint8_t buf8[128];

	if(ConsoleScanCmd("chd %127[^\n]", buf8) == 1)
	{
		f_chdrive((TCHAR*)buf8);

		f_getcwd((TCHAR*)buf8, bufLen);

		ConsolePrint("%s\r\n", buf8);
	}

	if(ConsoleCmpCmd("pwd"))
	{
		f_getcwd((TCHAR*)buf8, bufLen);

		ConsolePrint("%s\r\n", buf8);
	}

	if(ConsoleScanCmd("stat %127[^\n]", buf8) == 1)
	{
		FILINFO info;
		fresult = f_stat((TCHAR*)buf8, &info);

		if(fresult)
		{
			ConsolePrint("Stat error %u\r\n", fresult);
		}
		else
		{
			ConsolePrint("Size: %u\r\n", info.fsize);
		}
	}

	if(ConsoleCmpCmd("ls"))
	{
		DIR dir;
		FILINFO finfo;
		finfo.lfname = (TCHAR*)buf8;
		finfo.lfsize = bufLen-1;

		f_getcwd((TCHAR*)buf8, bufLen);

		ConsolePrint("Content of: %s\r\n", buf8);

		// show folders
		fresult = f_opendir(&dir, "");
		if(fresult != FR_OK)
			return;

		do
		{
			fresult = f_readdir(&dir, &finfo);

			if(finfo.fname[0] == 0)
				break;

			if(finfo.fattrib & AM_DIR)
			{
				ConsolePrint("<DIR>      %s\r\n", *finfo.lfname ? finfo.lfname : finfo.fname);
			}
		}
		while(fresult == FR_OK);

		f_closedir(&dir);

		// show files
		fresult = f_opendir(&dir, "");
		if(fresult != FR_OK)
			return;

		do
		{
			fresult = f_readdir(&dir, &finfo);

			if(finfo.fname[0] == 0)
				break;

			if((finfo.fattrib & AM_DIR) == 0)
			{
				if(finfo.fsize > 1024*1024*16)	// larger than 16M?
					ConsolePrint("%7uM   %s\r\n", finfo.fsize/(1024*1024), *finfo.lfname ? finfo.lfname : finfo.fname);
				else if(finfo.fsize > 1024*16)	// larget than 16k?
					ConsolePrint("%7uk   %s\r\n", finfo.fsize/1024, *finfo.lfname ? finfo.lfname : finfo.fname);
				else
					ConsolePrint("%7u    %s\r\n", finfo.fsize, *finfo.lfname ? finfo.lfname : finfo.fname);
			}
		}
		while(fresult == FR_OK);

		f_closedir(&dir);
	}

	if(ConsoleScanCmd("cd %127[^\n]", buf8) == 1)
	{
		f_chdir((TCHAR*)buf8);

		f_getcwd((TCHAR*)buf8, bufLen);

		ConsolePrint("%s\r\n", buf8);
	}

	if(ConsoleScanCmd("touch %127[^\n]", buf8) == 1)
	{
		FIL file;
		fresult = f_open(&file, (TCHAR*)buf8, FA_CREATE_NEW);
		if(fresult == FR_EXIST)
			ConsolePrint("File already exists\r\n");
		f_close(&file);
	}

	if(ConsoleScanCmd("mkdir %127[^\n]", buf8) == 1)
	{
		fresult = f_mkdir((TCHAR*)buf8);
		if(fresult == FR_OK)
			ConsolePrint("Created dir: %s\r\n", buf8);
	}

	if(ConsoleScanCmd("rm %127[^\n]", buf8) == 1)
	{
		fresult = f_unlink((TCHAR*)buf8);
		if(fresult != FR_OK)
			ConsolePrint("Could not remove object %d\r\n", fresult);
	}

	if(ConsoleScanCmd("logfile %127[^\n]", buf8) == 1)
	{
		LogFileClose();

		uint32_t freeCluster;
		FATFS* fs;
		fresult = f_getfree("SD:", &freeCluster, &fs);
		if(fresult)
		{
			NetworkPrint("No media attached\r\n");
			ConsolePrint("No media attached\r\n");
			return;
		}

		uint32_t freeSectors = freeCluster * fs->csize;
		uint32_t freeSizeMB = freeSectors/((1024*1024)/512);

		NetworkPrint("Free Space: %uMB\n", freeSizeMB);

		f_chdrive("SD:");
		LogFileOpen((char*)buf8);
	}

	if(ConsoleCmpCmd("startlog"))
	{
		ConsolePrint("Writing log header\r\n");

		LogFileOpen(0);
	}

	if(ConsoleCmpCmd("stoplog"))
	{
		if(logFile.ready == 0)
		{
			NetworkPrint("No log file open!\r\n");
			ConsolePrint("No log file open!\r\n");
		}
		else
		{
			LogFileClose();
		}
	}

	if(ConsoleCmpCmd("sd write"))
	{
		ConsolePrint("SD card write test\r\n");

//		static FIL file __attribute__((section(".dtc")));
		static FIL file;

		fresult = f_open(&file, "SD:/wrtest.txt", FA_CREATE_ALWAYS | FA_WRITE);

//		static uint8_t wrBuf[8192] __attribute__((section(".dtc")));
		static uint8_t wrBuf[8192];
		for(uint32_t i = 0; i < 8192; i++)
			wrBuf[i] = i;

		for(uint32_t i = 0; i < 64; i++)
		{
			UINT bytesWritten = 0;
			fresult = f_write(&file, wrBuf, 8001, &bytesWritten);
			ConsolePrint("Result: %d, %u bytes written\r\n", fresult, bytesWritten);

			chThdSleepMilliseconds(10);
		}

		f_close(&file);
	}
}

static void buzzerCommands()
{
	uint16_t u16;

	if(ConsoleCmpCmd("buzz off"))
	{
		ConsolePrint("Disabling buzzer\r\n");

		BuzzerPlay(0);
	}

	if(ConsoleScanCmd("buzz %hu", &u16) == 1)
	{
		ConsolePrint("Buzzer freq: %hu\r\n", u16);

		BuzzerTone(u16);
	}

	if(ConsoleCmpCmd("buzz up20"))
	{
		BuzzerPlay(&buzzSeqUp20);
	}

	if(ConsoleCmpCmd("buzz up50"))
	{
		BuzzerPlay(&buzzSeqUp50);
	}

	if(ConsoleCmpCmd("buzz down50"))
	{
		BuzzerPlay(&buzzSeqDown50);
	}

	if(ConsoleCmpCmd("buzz up100"))
	{
		BuzzerPlay(&buzzSeqUp100);
	}

	if(ConsoleCmpCmd("buzz down100"))
	{
		BuzzerPlay(&buzzSeqDown100);
	}

	if(ConsoleCmpCmd("buzz updown20"))
	{
		BuzzerPlay(&buzzSeqUpDown20);
	}

	if(ConsoleCmpCmd("buzz beep"))
	{
		BuzzerPlay(&buzzSeqBeepFast);
	}

	if(ConsoleCmpCmd("buzz doublebeep"))
	{
		BuzzerPlay(&buzzSeqDoubleBeepSlow);
	}

	if(ConsoleCmpCmd("buzz tada"))
	{
		BuzzerPlay(&buzzSeqTada);
	}
}

static void processCommand()
{
	buzzerCommands();
#ifndef ENV_BOOT
	miscCommands();
	fsCommands();
	motorsCommands();
	kickerCommands();
	wifiCommands();
#endif

	if(ConsoleCmpCmd("version"))
	{
		ConsolePrint("Firmware Date: %s - %s\r\n", __DATE__, __TIME__);
	}

	if(ConsoleCmpCmd("gitref"))
	{
		ConsolePrint("GIT ref: %s\r\n", GIT_REF);
	}
}
