/*
 * adns9800.c
 *
 *  Created on: 30.03.2016
 *      Author: AndreR
 */

#include "adns9800.h"
#include "adns9800_srom_A6.h"
#include "util/init_hal.h"

uint8_t ADNSReadReg(ADNS9800* pAD, uint8_t reg)
{
	pAD->tx[0] = reg & 0x7F;

	SPISyncSplitInitTransfer(&pAD->slave);	// Pull CS low
	SPISyncSplitTransfer(&pAD->slave, pAD->tx, 0, 1); // send address, MSB = 0 => read
	chThdSleepMicroseconds(250);	// tSRAD
	SPISyncSplitTransfer(&pAD->slave, 0, pAD->rx, 1); // read reg
	chThdSleepMicroseconds(250);; // tSCLK-NCS
	SPISyncSplitFinish(&pAD->slave); // CS high
	chThdSleepMicroseconds(250);; // tSRW/tSRR (=20us) minus tSCLK-NCS

	return pAD->rx[0];
}

void ADNSWriteReg(ADNS9800* pAD, uint8_t reg, uint8_t data)
{
	pAD->tx[0] = reg | 0x80;
	pAD->tx[1] = data;

	SPISyncSplitInitTransfer(&pAD->slave);	// Pull CS low
	SPISyncSplitTransfer(&pAD->slave, pAD->tx, 0, 2); // send address, MSB = 1 => write
	chThdSleepMicroseconds(250); // tSCLK-NCS
	SPISyncSplitFinish(&pAD->slave); // CS high
	chThdSleepMicroseconds(250); // tSRW/tSRR (=20us) minus tSCLK-NCS
}

void ADNS9800Init(ADNS9800* pAD, SPISync* pSPI, GPIO_TypeDef* pCSPort, uint32_t csPin)
{
	pAD->slave.cpha = 1;
	pAD->slave.cpol = 1;
	pAD->slave.csPin = csPin;
	pAD->slave.pCSPort = pCSPort;
	pAD->slave.timeoutTicks = MS2ST(20);
	pAD->slave.prescaler = SPI_CR1_BRDIV256;
	SPISyncSlaveInit(pSPI, &pAD->slave);
}

void ADNS9800UploadFirmware(ADNS9800* pAD)
{
	ADNSWriteReg(pAD, 0x3A, 0x5A);	// force reset

	chThdSleepMilliseconds(80);	// wait for reboot

	ADNSReadReg(pAD, 0x02);
	ADNSReadReg(pAD, 0x03);
	ADNSReadReg(pAD, 0x04);
	ADNSReadReg(pAD, 0x05);
	ADNSReadReg(pAD, 0x06);

	uint8_t prodId;
	uint8_t revId;
	ADNS9800GetProdRev(pAD, &prodId, &revId);

	if(prodId != 0x33)
	{
		pAD->functional = 0;
		return;
	}

	ADNSWriteReg(pAD, 0x39, 0x02);	// select 3k SROM
	ADNSWriteReg(pAD, 0x13, 0x1D);	// init SROM download
	chThdSleepMilliseconds(10);
	ADNSWriteReg(pAD, 0x13, 0x18);	// start SROM download

	SPISyncSplitInitTransfer(&pAD->slave);
	pAD->tx[0] = 0x62 | 0x80;	// SROM load burst register
	SPISyncSplitTransfer(&pAD->slave, pAD->tx, 0, 1);
	chThdSleepMicroseconds(250);
	for(uint16_t i = 0; i < sizeof(ADNS9800_SROM_A6); i++)
	{
		pAD->tx[0] = ADNS9800_SROM_A6[i];
		SPISyncSplitTransfer(&pAD->slave, pAD->tx, 0, 1);
		chThdSleepMicroseconds(250);
	}
	SPISyncSplitFinish(&pAD->slave);

	chThdSleepMilliseconds(10);

	uint8_t laserCtrl = ADNSReadReg(pAD, 0x20);
	ADNSWriteReg(pAD, 0x20, laserCtrl & 0xF0);

	ADNSWriteReg(pAD, 0x2E, 0x1F);
	ADNSWriteReg(pAD, 0x0F, 0x29);

	chThdSleepMilliseconds(5);

	pAD->functional = 1;
}

void ADNS9800SetResolution(ADNS9800* pAD, uint8_t val)
{
	ADNSWriteReg(pAD, 0x0F, val);
}

void ADNS9800GetProdRev(ADNS9800* pAD, uint8_t* pProdId, uint8_t* pRevId)
{
	*pProdId = ADNSReadReg(pAD, 0x00);
	*pRevId = ADNSReadReg(pAD, 0x01);
}

void ADNS9800UpdateMotion(ADNS9800* pAD, float tSample)
{
	pAD->tx[0] = 0x50;

	SPISyncSplitInitTransfer(&pAD->slave);	// Pull CS low
	SPISyncSplitTransfer(&pAD->slave, pAD->tx, 0, 1); // send address, MSB = 0 => read
	chThdSleepMicroseconds(250);	// tSRAD
	SPISyncSplitTransfer(&pAD->slave, 0, pAD->rx, 14); // read reg
	chThdSleepMicroseconds(250); // tSCLK-NCS
	SPISyncSplitFinish(&pAD->slave); // CS high
	chThdSleepMicroseconds(250); // tSRW/tSRR (=20us) minus tSCLK-NCS

	pAD->motion.motion = pAD->rx[0];
	pAD->motion.deltaX = pAD->rx[2] | ((uint16_t)pAD->rx[3]) << 8;
	pAD->motion.deltaY = pAD->rx[4] | ((uint16_t)pAD->rx[5]) << 8;
	pAD->motion.squal = pAD->rx[6];
	pAD->motion.pixelSum = pAD->rx[7];
	pAD->motion.maxPixel = pAD->rx[8];
	pAD->motion.minPixel = pAD->rx[9];
	pAD->motion.shutter = ((uint16_t)pAD->rx[10]) << 8 | pAD->rx[11];
	pAD->motion.framePeriod = ((uint16_t)pAD->rx[12]) << 8 | pAD->rx[13];

	pAD->motion.sumX += pAD->motion.deltaX;
	pAD->motion.sumY += pAD->motion.deltaY;

	pAD->motion.speed[0] = ((float)pAD->motion.deltaX)/8200.0f*0.0254f*1.0f/tSample*1.0f;
	pAD->motion.speed[1] = ((float)pAD->motion.deltaY)/8200.0f*0.0254f*1.0f/tSample*1.0f;
}
