/*
 * buzzer.c
 *
 *  Created on: 23.12.2010
 *      Author: AndreR
 */

#include "buzzer.h"

#include "stm32f746xx.h"
#include "util/init_hal.h"

Buzzer buzzer;

const BuzzerSequence buzzSeqBeepFast = {
		(uint16_t[]){1, 100, 0, 100},
		4, 0,
};

const BuzzerSequence buzzSeqDoubleBeepSlow = {
		(uint16_t[]){1, 100, 0, 100, 1, 100, 0, 2000},
		8, 0,
};

const BuzzerSequence buzzSeqUp100 = {
		(uint16_t[]){2800, 100, 3000, 100, 3200, 100},
		6, 1,
};

const BuzzerSequence buzzSeqTada = {
		(uint16_t[]){3200, 100, 2900, 100, 2600, 100, 3200, 300, 2600, 100, 3200, 200},
		12, 1,
};

const BuzzerSequence buzzSeqDown100 = {
		(uint16_t[]){3200, 100, 3000, 100, 2800, 100},
		6, 1,
};

static uint16_t buzzSeqUp50Data[32];

const BuzzerSequence buzzSeqUp50 = {
		buzzSeqUp50Data, 32, 1,
};

static uint16_t buzzSeqDown50Data[32];

const BuzzerSequence buzzSeqDown50 = {
		buzzSeqDown50Data, 32, 1,
};

static uint16_t buzzSeqUp20Data[32];

const BuzzerSequence buzzSeqUp20 = {
		buzzSeqUp20Data, 32, 1,
};

static uint16_t buzzSeqUpDown20Data[64];

const BuzzerSequence buzzSeqUpDown20 = {
		buzzSeqUpDown20Data, 64, 0,
};

#define BUZZER_ON() (TIM10->CCMR1 = TIM_CCMR1_OC1M_0 | TIM_CCMR1_OC1M_1 | TIM_CCMR1_OC1M_2)
#define BUZZER_OFF() (TIM10->CCMR1 = TIM_CCMR1_OC1M_0 | TIM_CCMR1_OC1M_2)

void BuzzerInit()
{
	RCC->APB2ENR |= RCC_APB2ENR_TIM10EN;	// enable TIM10 clock

	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 3;
	gpioInit.ospeed = GPIO_OSPEED_2MHZ;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOB, GPIO_PIN_8, &gpioInit);

	TIM10->CR1 = TIM_CR1_ARPE;	// enable ARR preload
	BUZZER_OFF();
	TIM10->CCER = TIM_CCER_CC1E | TIM_CCER_CC1P;	// enable CC1 output and set polarity to active low
	TIM10->PSC = 215;		// prescaler
	TIM10->ARR = 320;	// period
	TIM10->CCR1 = 103;	// pulse
	TIM10->EGR = TIM_EGR_UG;	// generate update event to take new ARR value
	TIM10->CR1 |= TIM_CR1_CEN;	// enable timer

	// initalize buzzer sequences
	for(uint16_t i = 0; i < 32; i += 2)
	{
		buzzSeqUp50Data[i] = 2600+i*25;
		buzzSeqUp50Data[i+1] = 50;

		buzzSeqUp20Data[i] = 2600+i*25;
		buzzSeqUp20Data[i+1] = 20;

		buzzSeqUpDown20Data[i] = 2600+i*25;
		buzzSeqUpDown20Data[i+1] = 20;
	}

	for(uint16_t i = 32; i < 64; i += 2)
	{
		buzzSeqUpDown20Data[i] = 3400-(i-32)*25;
		buzzSeqUpDown20Data[i+1] = 20;

		buzzSeqDown50Data[i-32] = 3400-(i-32)*25;
		buzzSeqDown50Data[i-31] = 50;
	}
}

void BuzzerTone(uint16_t freq)
{
	if(freq == 1)
		freq = 3150;

	if(freq < 16 || freq > 20000)
	{
		BUZZER_OFF();
		return;
	}

	BUZZER_ON();

	uint16_t arr = 1000000/freq;

	TIM10->ARR = arr;
	TIM10->CCR1 = arr/2;
	TIM10->EGR = TIM_EGR_UG;
}

static void buzzTimer(void* param)
{
	(void)param;

	if(!buzzer.pSeq)
		return;

	if(buzzer.curTone == 0 && buzzer.pSeq->once)
	{
		buzzer.pSeq = 0;
		BuzzerTone(0);
		return;
	}

	BuzzerTone(buzzer.pSeq->pSong[buzzer.curTone]);

	chSysLockFromISR();
	chVTSetI(&buzzer.timer, MS2ST(buzzer.pSeq->pSong[buzzer.curTone+1]), &buzzTimer, 0);
	chSysUnlockFromISR();

	buzzer.curTone += 2;
	buzzer.curTone %= buzzer.pSeq->numTones;
}

void BuzzerPlay(const BuzzerSequence* pSeq)
{
	if(pSeq == buzzer.pSeq && pSeq != 0)
		return;

	chVTReset(&buzzer.timer);
	buzzer.pSeq = 0;
	BuzzerTone(0);

	if(pSeq == 0)
		return;

	if(pSeq->numTones % 2 == 1 || pSeq->numTones == 0)
		return;

	buzzer.curTone = 2;
	buzzer.pSeq = pSeq;

	BuzzerTone(buzzer.pSeq->pSong[0]);
	chVTSet(&buzzer.timer, MS2ST(buzzer.pSeq->pSong[1]), &buzzTimer, 0);
}
