/*
 * network_impl.c
 *
 *  Created on: 01.07.2010
 *      Author: AndreR
 */

#include "network_impl.h"

#include "stm32f746xx.h"
#include "constants.h"

#include "main/network.h"

#include "util/console.h"
#include "util/init_hal.h"
#include "util/intercom.h"

#include "hal/buzzer.h"
#include "hal/kicker.h"
#include "hal/motors.h"

#include "config_main.h"

#include <string.h>

NetworkImpl networkImpl;

void DMA1_Stream3_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	SPISyncDmaRxInterrupt(&network.nrf24.spi);

	CH_IRQ_EPILOGUE();
}

void EXTI1_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = (1 << 1);

	NRF24ExIRQ(&network.nrf24);

	CH_IRQ_EPILOGUE();
}

void TIM7_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM7->SR = 0;
	TIM7->CR1 = 0;

	chSysLockFromISR();
	chBSemSignalI(&network.timerSem);
	chSysUnlockFromISR();

	CH_IRQ_EPILOGUE();
}

void NetworkImplStartTimeout(uint32_t microSeconds)
{
	TIM7->CNT = 0;
	TIM7->ARR = microSeconds;
	TIM7->CR1 = TIM_CR1_CEN/* | TIM_CR1_OPM*/;
}

static void networkIntercomCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command == IC_NETWORK_NORMAL)
		NetworkSendPacket((PacketHeader*)pData, pData+PACKET_HEADER_SIZE, dataLength-PACKET_HEADER_SIZE);

	if(pHeader->command == IC_NETWORK_RELIABLE)
		NetworkSendPacketReliable((PacketHeader*)pData, pData+PACKET_HEADER_SIZE, dataLength-PACKET_HEADER_SIZE);
}

void NetworkImplInit()
{
	// NRF24 IRQ line setup
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.pupd = GPIO_PUPD_UP;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_FALLING;
	GPIOInit(GPIOB, GPIO_PIN_1, &gpioInit);

	NVICEnableIRQ(EXTI1_IRQn, IRQL_NRF24);

	// TIM7 setup
	RCC->APB1ENR |= RCC_APB1ENR_TIM7EN;

	TIM7->CR1 = 0;
	TIM7->CR2 = 0;
	TIM7->SR = 0;
	TIM7->DIER = TIM_DIER_UIE;
	TIM7->PSC = 215; 	// clock is now 1MHz
	TIM7->ARR = 1000;	// = 1ms

	NVICEnableIRQ(TIM7_IRQn, IRQL_NRF24);

	// SPI2 init
	RCC->APB1ENR |= RCC_APB1ENR_SPI2EN;

	NVICEnableIRQ(DMA1_Stream3_IRQn, IRQL_NRF24);

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 5;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_25MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOC, GPIO_PIN_2 | GPIO_PIN_3, &gpioInit);
	GPIOInit(GPIOD, GPIO_PIN_3, &gpioInit);

	// NRF24 setup
	NRF24ExData data;

	static uint8_t rxData[NRF24_RX_SIZE];
	static uint8_t txData[NRF24_TX_SIZE];

	NRF24QueueInit(&network.queue, txData, NRF24_TX_SIZE, rxData, NRF24_RX_SIZE);
	network.queue.id = network.config.botId;

	data.csPin = GPIO_PIN_4;
	data.pCSPort = GPIOA;
	data.spiData.pRegister = SPI2;
	data.spiData.pDma = DMA1;
	data.spiData.dmaChRx = 3;
	data.spiData.dmaChTx = 4;
	data.rx = 1;
	data.channel = network.config.channel;
	data.speed = network.config.speed;
	data.pCEPort = GPIOF;
	data.cePin = GPIO_PIN_11;

	NRF24ExInit(&network.nrf24, &data);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &networkIntercomCallback;
	client.section = INTERCOM_SECTION_NETWORK;

	IntercomAddClient(client);
}

void NetworkImplWifiStatsUpdate(PresenterWifiStat* pStats)
{
	memcpy(&networkImpl.wifiStats, pStats, sizeof(PresenterWifiStat));
}

// this is the main network processing function
void NetworkImplProcessPacket(PacketHeader* pHeader, uint8_t* pBuf, uint16_t dataLength)
{
	switch(pHeader->section)
	{
		case SECTION_SYSTEM:
		{
			switch(pHeader->cmd)
			{
				case CMD_SYSTEM_CONSOLE_COMMAND:
				{
//					SystemConsoleCommand* pCmd = (SystemConsoleCommand*)pBuf;

//					if(pCmd->target == INTERCOM_ADDRESS_MAIN)
					{
						// process onboard
						uint8_t* pData;
						uint32_t length = dataLength-sizeof(SystemConsoleCommand);
						int16_t result = FifoLinReserve(&console.rxFifo, length, &pData);
						if(result)
							return;

						memcpy(pData, pBuf+sizeof(SystemConsoleCommand), length);

						FifoLinCommit(&console.rxFifo, length);
					}
//					else
//					{
//						if(cli.intercomFwd.inUse == 0)
//						{
//							cli.intercomFwd.bufSize = dataLength-sizeof(SystemConsoleCommand);
//
//							if(cli.intercomFwd.bufSize <= CLI_INTERCOM_BUF_SIZE)
//							{
//								cli.intercomFwd.inUse = 1;
//								cli.intercomFwd.target = pCmd->target;
//
//								memcpy(cli.intercomFwd.buf, pBuf+sizeof(SystemConsoleCommand), cli.intercomFwd.bufSize);
//							}
//						}
//					}
				}
				break;
				case CMD_SYSTEM_MATCH_CTRL:
				{
					SystemMatchCtrl* pCtrl = (SystemMatchCtrl*)pBuf;

					if(pCtrl->flags & SYSTEM_MATCH_CTRL_FLAGS_CHEER)
					{
						BuzzerPlay(&buzzSeqUpDown20);
					}
					else
					{
						if(buzzer.pSeq == &buzzSeqUpDown20)
						{
							BuzzerPlay(0);
						}
					}

					// convert Q3.5 to float
					float kickSpeed = (pCtrl->kickSpeed & 0x1F)*0.032f + ((pCtrl->kickSpeed & 0xE0) >> 5);

					KdCommand kd;
					kd.autoChg = pCtrl->kickFlags & SYSTEM_MATCH_CTRL_KICK_FLAGS_AUTOCHG_MASK;
					kd.device = (pCtrl->kickFlags & SYSTEM_MATCH_CTRL_KICK_FLAGS_DEVICE_MASK);
					kd.mode = (pCtrl->kickFlags & SYSTEM_MATCH_CTRL_KICK_FLAGS_MODE_MASK) >> 1;
					kd.speed = kickSpeed;
					kd.dribblerSpeed = pCtrl->dribblerSpeed*100.0f;

					// Kicker
					if(kd.autoChg)
					{
						KickerEnableAutoRecharge(1);
						KickerEnableCharge(1);
					}
					else
					{
						KickerAutoDischarge();
					}

					switch(kd.mode)
					{
						case KICKER_MODE_FORCE:
						{
							KickerFireSpeed(kd.speed, kd.device);
//							ConsolePrint("Fire %hu: %.3f\r\n", (uint16_t)kd.device, kd.duration);
						}
						break;
						case KICKER_MODE_ARM:
						{
							KickerArm(kd.speed, kd.device);
						}
						break;
						case KICKER_MODE_AIM_ARM:
						{
							KickerAimArm(kd.speed, kd.device);
						}
						break;
						default: KickerDisarm(); break;
					}

					MotorsDribblerSetVelocity(kd.dribblerSpeed);
				}
				break;
			}
		}
		break;
		case SECTION_BOOTLOADER:
		{
//			if(pHeader->cmd == CMD_BOOTLOADER_CHECK)
//			{
//				IntercomHeader iHeader;
//				iHeader.address = INTERCOM_ADDRESS_MEDIA;
//				iHeader.section = INTERCOM_SECTION_NETWORK;
//				iHeader.command = 1;
//
//				IntercomReliableSend(&iHeader, 0, 0);
//			}
//			else
//			{
//				IntercomHeader iHeader;
//				iHeader.address = INTERCOM_ADDRESS_MEDIA;
//				iHeader.section = INTERCOM_SECTION_NETWORK;
//				iHeader.command = 0;
//
//				IntercomSend(&iHeader, dataLength+PACKET_HEADER_SIZE, pHeader);
//			}
		}
		break;

#ifdef ENV_RUN
		case SECTION_CONFIG:
		{
			ConfigNetworkInput(pHeader, pBuf, dataLength);
		}
		break;
#endif
		default: break;
	}
}
