/*
 * spi1.c
 *
 *  Created on: 16.10.2015
 *      Author: AndreR
 */

#include "spi1.h"

#include "constants.h"

#include "util/console.h"
#include "util/init_hal.h"
#include "util/intercom.h"
#include "util/sys_time.h"
#include "hal/exti.h"
#include "hal/adns9800_srom_A6.h"
#include "gui/optical.h"

SPI1Global spi1;

#define EVENT_MOUSE_RDY		0x01

#define STATE_IDLE			0
#define STATE_ACTIVE		1

#define DMA_Stream_RX		DMA2_Stream2
#define DMA_Stream_TX		DMA2_Stream3

static uint8_t tx[10];
static uint8_t rx[20];

void DMA2_Stream2_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	SPISyncDmaRxInterrupt(&spi1.bus);

	CH_IRQ_EPILOGUE();
}

static void mouseIRQ()
{
	chSysLockFromISR();
	chMBPostI(&spi1.eventQueue, EVENT_MOUSE_RDY);
	chSysUnlockFromISR();
}

void SPI1Init()
{
	// init variables
	chMBObjectInit(&spi1.eventQueue, spi1.eventQueueData, SPI1_EVENT_QUEUE_SIZE);

	RCC->APB2ENR |= RCC_APB2ENR_SPI1EN;

	SPISyncData spiData;
	spiData.dmaChRx = 2;
	spiData.dmaChTx = 3;
	spiData.dmaPrio = 1;
	spiData.pDma = DMA2;
	spiData.pRegister = SPI1;
	SPISyncInit(&spi1.bus, &spiData);

	spi1.slave.cpha = 1;
	spi1.slave.cpol = 1;
	spi1.slave.csPin = GPIO_PIN_4;
	spi1.slave.pCSPort = GPIOA;
	spi1.slave.timeoutTicks = MS2ST(20);
	spi1.slave.prescaler = SPI_CR1_BRDIV256;
	SPISyncSlaveInit(&spi1.bus, &spi1.slave);

#ifdef DEBUG
	spi1.pRxDma = DMA_Stream_RX;
	spi1.pTxDma = DMA_Stream_TX;
	spi1.pSPI = SPI1;
	spi1.pDMA = DMA2;
#endif

	// init GPIO
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 5;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_25MHZ;
	gpioInit.pupd = GPIO_PUPD_NONE;
	GPIOInit(GPIOA, GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7, &gpioInit); // SCK, MISO, MOSI

	// configure NVIC
	NVICEnableIRQ(DMA2_Stream2_IRQn, IRQL_SPI1);

	// DRDY GYR IRQ
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_FALLING;
	gpioInit.pupd = GPIO_PUPD_UP;
//	GPIOInit(GPIOA, GPIO_PIN_9, &gpioInit);

//	ExtiSetCallback(&mouseIRQ, 9);
}

uint8_t ADNSReadReg(uint8_t reg)
{
	uint8_t result = 0;
	reg &= 0x7F;

	SPISyncSplitInitTransfer(&spi1.slave);	// Pull CS low
	SPISyncSplitTransfer(&spi1.slave, &reg, 0, 1); // send address, MSB = 0 => read
	chThdSleepMilliseconds(1);	// tSRAD
	SPISyncSplitTransfer(&spi1.slave, 0, &result, 1); // read reg
	chThdSleepMilliseconds(1); // tSCLK-NCS
	SPISyncSplitFinish(&spi1.slave); // CS high
	chThdSleepMilliseconds(1); // tSRW/tSRR (=20us) minus tSCLK-NCS

	return result;
}

void ADNSWriteReg(uint8_t reg, uint8_t data)
{
	tx[0] = reg | 0x80;
	tx[1] = data;

	SPISyncSplitInitTransfer(&spi1.slave);	// Pull CS low
	SPISyncSplitTransfer(&spi1.slave, tx, 0, 2); // send address, MSB = 1 => write
	chThdSleepMilliseconds(1); // tSCLK-NCS
	SPISyncSplitFinish(&spi1.slave); // CS high
	chThdSleepMilliseconds(1); // tSRW/tSRR (=20us) minus tSCLK-NCS
}

void ADNS9800Init()
{
	ADNSWriteReg(0x3A, 0x5A);	// force reset

	chThdSleepMilliseconds(80);	// wait for reboot

	ADNSReadReg(0x02);
	ADNSReadReg(0x03);
	ADNSReadReg(0x04);
	ADNSReadReg(0x05);
	ADNSReadReg(0x06);

	ADNSWriteReg(0x39, 0x02);	// select 3k SROM
	ADNSWriteReg(0x13, 0x1D);	// init SROM download
	chThdSleepMilliseconds(10);
	ADNSWriteReg(0x13, 0x18);	// start SROM download

	SPISyncSplitInitTransfer(&spi1.slave);
	tx[0] = 0x62 | 0x80;	// SROM load burst register
	SPISyncSplitTransfer(&spi1.slave, tx, 0, 1);
	chThdSleepMilliseconds(1);
	for(uint16_t i = 0; i < sizeof(ADNS9800_SROM_A6); i++)
	{
		SPISyncSplitTransfer(&spi1.slave, &ADNS9800_SROM_A6[i], 0, 1);
		chThdSleepMilliseconds(1);
	}
	SPISyncSplitFinish(&spi1.slave);

	chThdSleepMilliseconds(10);

	uint8_t laserCtrl = ADNSReadReg(0x20);
	ADNSWriteReg(0x20, laserCtrl & 0xF0);

	ADNSWriteReg(0x2E, 0x1F);

	ADNSWriteReg(0x0F, 0x29);

	chThdSleepMilliseconds(5);
}

void ADNS9800SetResolution(uint8_t val)
{
	ADNSWriteReg(0x0F, val);
}

void ADNS9800GetProdRev(uint8_t* pProdId, uint8_t* pRevId)
{
	*pProdId = ADNSReadReg(0x00);
	*pRevId = ADNSReadReg(0x01);
}

void ADNS9800GetMotion(ADNS9800Motion* pMot)
{
	uint8_t reg = 0x50;

	SPISyncSplitInitTransfer(&spi1.slave);	// Pull CS low
	SPISyncSplitTransfer(&spi1.slave, &reg, 0, 1); // send address, MSB = 0 => read
	chThdSleepMilliseconds(1);	// tSRAD
	SPISyncSplitTransfer(&spi1.slave, 0, rx, 14); // read reg
	chThdSleepMilliseconds(1); // tSCLK-NCS
	SPISyncSplitFinish(&spi1.slave); // CS high
	chThdSleepMilliseconds(1); // tSRW/tSRR (=20us) minus tSCLK-NCS

//	for(uint8_t i = 0; i < 14; i++)
//	{
//		ConsolePrint("%hu: %hu\r\n", (uint16_t)i, (uint16_t)rx[i]);
//	}

	pMot->motion = rx[0];
	pMot->deltaX = rx[2] | ((uint16_t)rx[3]) << 8;
	pMot->deltaY = rx[4] | ((uint16_t)rx[5]) << 8;
	pMot->squal = rx[6];
	pMot->pixelSum = rx[7];
	pMot->maxPixel = rx[8];
	pMot->minPixel = rx[9];
	pMot->shutter = ((uint16_t)rx[10]) << 8 | rx[11];
	pMot->framePeriod = ((uint16_t)rx[12]) << 8 | rx[13];

//	ConsolePrint("Motion: 0x%02hX\r\n", (uint16_t)mot.motion);
//	ConsolePrint("dX: %hd\r\ndY: %hd\r\n", mot.deltaX, mot.deltaY);
//	ConsolePrint("SQUAL: %hu\r\n", (uint16_t)mot.squal);
//	ConsolePrint("pixelSum: %hu\r\n", (uint16_t)mot.pixelSum);
//	ConsolePrint("maxPixel: %hu\r\n", (uint16_t)mot.maxPixel);
//	ConsolePrint("minPixel: %hu\r\n", (uint16_t)mot.minPixel);
//	ConsolePrint("Shutter: %hu\r\nFrame Period: %hu\r\n", mot.shutter, mot.framePeriod);
}

void SPI1Task(void* params)
{
	(void)params;
	ADNS9800Motion mot;

	chRegSetThreadName("SPI1");

	chThdSleepMilliseconds(100);

	ADNS9800Init();

	msg_t event;
	while(1)
	{
		ADNS9800GetMotion(&mot);

		OpticalUpdate(&mot);

		chThdSleepMilliseconds(20);

//		msg_t chResult = chMBFetch(&spi1.eventQueue, &event, MS2ST(10));
//		if(chResult == RDY_OK)
//		{
//			if(event & EVENT_MOUSE_RDY)
//			{
//			}
//		}
//		else
//		{
//		}
	}
}
