/*
 * log_msgs.h
 *
 *  Created on: 21.10.2014
 *      Author: AndreR
 */

#ifndef LOG_MSGS_H_
#define LOG_MSGS_H_

#include <stdint.h>
#include "util/pid.h"

#ifndef PACKED
#define PACKED __attribute__((packed))
#endif

// fundamental header for all entries in the log file
typedef struct PACKED _LogChunkHeader
{
	uint16_t type;
	uint16_t length;
} LogChunkHeader;

// the header for all log entries, contains the LogChunkHeader
typedef struct PACKED _LogEntryHeader
{
	uint16_t type;
	uint16_t length;
	uint32_t timestamp;
} LogEntryHeader;

// --- LOG DEFINITIONS ---
typedef struct PACKED _LogRawI16XYZ
{
	LogEntryHeader header;
	int16_t xyz[3];
} LogRawI16XYZ;

#define MOT_FR 0
#define MOT_FL 1
#define MOT_RL 2
#define MOT_RR 3

typedef struct PACKED _LogCtrlInput
{
	LogEntryHeader header;

	struct PACKED
	{
		uint32_t time;
		float vel[4];		// angular motor velocity [rad/s]
	} enc;

	struct PACKED
	{
		uint8_t updated;
		uint32_t time;
		float rotVel[3];	// around X, Y, Z [rad/s]
	} gyr;

	struct PACKED
	{
		uint8_t updated;
		uint32_t time;
		float linAcc[3];	// X, Y, Z [m^2/s]
	} acc;

	struct PACKED
	{
		uint8_t updated;
		uint32_t time;		// local arrival time (systime)
		uint32_t delay;		// t [us]
		float pos[3];		// X, Y, Z [m]
		uint8_t camId;
	} vision;

	struct PACKED
	{
		float voltage;		// battery voltage [V]
		float current;		// main current consumption [A]
		uint8_t batEmpty;
	} power;

	struct PACKED
	{
		uint32_t time;
		float vol[4];		// [V]
	} motorVol;
} LogCtrlInput;

typedef struct PACKED _LogCtrlState
{
	LogEntryHeader header;

	float pos[3];	// [m]
	float vel[3];	// [m/s]
	float acc[3];

	uint8_t posUpdated;
	uint8_t velUpdated;
	uint8_t accUpdated;
} LogCtrlState;

typedef struct PACKED _LogCtrlOutput
{
	LogEntryHeader header;

	float motorVel[4];	// angular motor velocity [rad/s]
	float motorVoltage[4];	// Motor voltage [V], -CCW, +CW
	uint8_t ctrlMode;
} LogCtrlOutput;

typedef struct PACKED _LogCtrlTiggaDebug
{
	LogEntryHeader header;

	float kfSigmaPos[3];
	float kfSigmaVel[3];
	float kfSigmaAcc[3];

	uint8_t ctrlMode;
	float setPos[3];
	float setVel[3];

	float trajPos[3];
	float trajVelLocal[3];
	float trajAccLocal[3];
	float trajJerkLocal[3];

	uint32_t fusionTime;
	uint32_t ctrlTime;

	float enc2Real[2];
} LogCtrlTiggaDebug;

typedef struct PACKED _LogKickerDribbler
{
	LogEntryHeader header;

	float kickerLevel;		// [V]
	float kickerChgCurrent; // [A]
	uint8_t kickCounter;

	int16_t dribblerSpeed; 	// [RPM]
	float dribblerTemp; 	// [C]
	uint8_t dribblerOverheated;

	uint16_t irLevel[2];
	uint8_t barrierInterrupted;
} LogKickerDribbler;

typedef struct PACKED _LogCtrlDriveTrain
{
	LogEntryHeader header;

	// drive train
	float motor2WheelRatio;
	float Km;	// motor torque constant [Nm/A]
	float Kn;	// motor speed constant [rad/(s*V)]
	float Ke;	// motor back-EMF constant [V/rad]
	float Kf;	// motor friction constant [Nm*s]
	float R;	// motor resistance [Ohm]
	float L;	// motor inductance [H]
} LogCtrlDriveTrain;

typedef struct PACKED _LogOpticalRaw
{
	LogEntryHeader header;

	int16_t lDeltaX;
	int16_t lDeltaY;
	uint8_t lSqual;
	int16_t rDeltaX;
	int16_t rDeltaY;
	uint8_t rSqual;
} LogOpticalRaw;

// LOGFILE descriptions
enum ElementType
{
	UINT8 = 0,
	INT8,
	UINT16,
	INT16,
	UINT32,
	INT32,
	FLOAT,
	DOUBLE,
};

// describes a single structure element
typedef struct PACKED _ElementDesc
{
	uint8_t type;
	const char* pName;	// 60 characters max
	const char* pUnit;
	const char* pDesc;
} ElementDesc;

// describes a message
typedef struct PACKED _LogMessageDesc
{
	uint16_t msgId;
	const char* pName;
	uint16_t numElements;
	ElementDesc* elements;
} LogMessageDesc;

extern const LogMessageDesc logMessageDescriptions[9];

#endif /* LOG_MSGS_H_ */
