/*
 * ctrl.c
 *
 *  Created on: 06.03.2013
 *      Author: AndreR
 */

#include <main/ctrl.h>
#include <main/skills.h>
#include "util/sys_time.h"
#include "struct_ids.h"
#include "util/arm_mat_util_f32.h"

#include <string.h>

Ctrl ctrl = {
	.botLogging = { 1, 1 },
	.misc = { 0xFF, 0xFF },
};

static const ConfigFileDesc configFileDescPhysical =
	{ SID_CFG_CTRL_PHYSICAL_PARAMS, 1, "ctrl/physical", 5, (ElementDesc[]) {
		{ FLOAT, "wheel_radius", "m", "Wheel Radius" },
		{ FLOAT, "front_angle", "deg", "Front Angle" },
		{ FLOAT, "back_angle", "deg", "Back  Angle" },
		{ FLOAT, "bot_radius", "m", "Bot Radius" },
		{ FLOAT, "mass", "kg", "Mass" },
	} };

static const ConfigFileDesc configFileDescLogging =
	{ SID_CFG_CTRL_LOGGING, 2, "ctrl/logs", 2, (ElementDesc[]) {
		{ UINT8, "in_state_out", "", "Input/State/Output" },
		{ UINT8, "specific", "", "Specific Ctrl. Data" },
	} };

static float pXYW2Motor[12];
static float pMotor2XYW[12];
static float pForceXYW2Motor[12];
static float pForceMotor2XYW[12];
static float pMotorNullDir[4];
static float pForceXYZ2DownForce[12];

static void updatePhysicalParameters();
static void updateParameters();

static void doControl();
static void doLogging();
static void clearUpdateFlags();
static void checkControllerChange();
static void sendMatchFeedback();
static void stateWatchdog();

static void configUpdateCallback(uint16_t cfgId)
{
	switch(cfgId)
	{
		case SID_CFG_CTRL_PHYSICAL_PARAMS:
		{
			updatePhysicalParameters();

			// reload controller so that its init function can take the new parameters into account
			CtrlSetController(ctrl.pInstance);	// FIXME: this does not have any effect
		}
		break;
	}
}

void CtrlInit()
{
	ctrl.logInterval = 100000;
	ctrl.buzzOnTimeViolation = 1;
	ctrl.ctrlIntervalUSec = 1000;

	ctrl.input.header.type = SID_CTRL_INPUT;
	ctrl.input.header.length = sizeof(LogCtrlInput);
	ctrl.state.header.type = SID_CTRL_STATE;
	ctrl.state.header.length = sizeof(LogCtrlState);
	ctrl.output.header.type = SID_CTRL_OUTPUT;
	ctrl.output.header.length = sizeof(LogCtrlOutput);
	ctrl.kd.header.type = SID_KICKER_DRIBBLER;
	ctrl.kd.header.length = sizeof(LogKickerDribbler);
	ctrl.logCfgDriveTrain.header.type = SID_CTRL_DRIVE_TRAIN;
	ctrl.logCfgDriveTrain.header.length = sizeof(LogCtrlDriveTrain);

	EMAFilterInit(&ctrl.avgCtrlRuntime, 0.96f);

	LagElementPT1Init(&ctrl.accZLag, 1, 0.5f, 0.001f);
	ctrl.accZLag.state[0] = 0.0f;
	ctrl.accZLag.state[1] = 0.0f;
	ctrl.accZLag.state[2] = 10.0f;
	ctrl.accZLag.state[3] = 10.0f;
	LagElementPT1Init(&ctrl.gyrWLag, 1, 0.5f, 0.001f);

	// construct matrices
	ctrl.matXYW2Motor = (arm_matrix_instance_f32){4, 3, pXYW2Motor};
	ctrl.matMotor2XYW = (arm_matrix_instance_f32){3, 4, pMotor2XYW};
	ctrl.matForceXYW2Motor = (arm_matrix_instance_f32){4, 3, pForceXYW2Motor};
	ctrl.matForceMotor2XYW = (arm_matrix_instance_f32){3, 4, pForceMotor2XYW};
	ctrl.matMotorNullDir = (arm_matrix_instance_f32){4, 1, pMotorNullDir};
	ctrl.matForceXYZ2DownForce = (arm_matrix_instance_f32){4, 3, pForceXYZ2DownForce};
	ctrl.matPos = (arm_matrix_instance_f32){3, 1, ctrl.state.pos};
	ctrl.matVel = (arm_matrix_instance_f32){3, 1, ctrl.state.vel};

	FlashFSOpenOrCreate("botStatus", 0, &ctrl.misc, sizeof(ConfigMisc), &ctrl.pMiscConfig);

	ctrl.pBotParamsConfig = ConfigOpenOrCreate(&configFileDescPhysical, &botParams.physical, sizeof(PhysicalParams), &configUpdateCallback);
	ctrl.pBotLoggingConfig = ConfigOpenOrCreate(&configFileDescLogging, &ctrl.botLogging, sizeof(BotLogging), &configUpdateCallback);

	ConfigNotifyUpdate(ctrl.pBotParamsConfig);

	memcpy(ctrl.input.vision.pos, ctrl.defaultPosition, 3*sizeof(float_t));

	CtrlImplInit();
}

void CtrlTask(void* params)
{
	(void) params;

	chRegSetThreadName("CTRL");

	systime_t prev = chVTGetSystemTimeX();
	systime_t next = prev+US2ST(ctrl.ctrlIntervalUSec);

	while(1)
	{
		// Log ctrl start time for runtime analysis
		uint32_t ctrlStartTime = SysTimeUSecRaw();

		ctrl.cpuUsage = ctrl.avgCtrlRuntime.value/((float)ctrl.ctrlIntervalUSec)*100.0f;

		// fetch updated control input data
		CtrlImplUpdateInput(&ctrl.input, &ctrl.kd);
		updateParameters();

		// check Z acc and angular speed (robot fell over? gyro damaged?)
		stateWatchdog();

		// primary control loop with sensor fusion
		doControl();

		// perform logging to SD card
		doLogging();

		// send Feedback to Sumatra
		sendMatchFeedback();

		// clear update flags on control state and input
		clearUpdateFlags();

		// check if a controller change is requested
		checkControllerChange();

		// Update average filter for runtime measurement
		EMAFilterUpdate(&ctrl.avgCtrlRuntime, SysTimeUSecRaw()-ctrlStartTime);

		// wait until next loop
		chThdSleepUntilWindowed(prev, next);
		prev = next;
		next += US2ST(ctrl.ctrlIntervalUSec);
	}
}

void CtrlUpdateBotId(uint8_t botId)
{
	CtrlUtilGetDefaultBotPosition(botId, ctrl.defaultPosition);
}

void CtrlSetHardwareId(uint8_t hwId)
{
	ctrl.misc.hardwareId = hwId;
	FlashFSWrite(&ctrl.pMiscConfig, &ctrl.misc, sizeof(ConfigMisc));
}

void CtrlSetFeatures(uint16_t features)
{
	ctrl.misc.features = features;
	FlashFSWrite(&ctrl.pMiscConfig, &ctrl.misc, sizeof(ConfigMisc));
}

void CtrlSetParams(float frontAngle, float backAngle, float botRadius,
		float wheelRadius, float mass)
{
	botParams.physical.frontAngle = frontAngle;
	botParams.physical.backAngle = backAngle;
	botParams.physical.botRadius = botRadius;
	botParams.physical.wheelRadius = wheelRadius;
	botParams.physical.mass = mass;

	ConfigNotifyUpdate(ctrl.pBotParamsConfig);
}

static void updatePhysicalParameters()
{
	arm_matrix_instance_f32 matCM2Wheel = {4, 3, (float[12]){}};
	const float centerOfMass[3] = {0, 0, 0.056};

	float frontAngle = botParams.physical.frontAngle;
	float backAngle = botParams.physical.backAngle;
	float botRadius = botParams.physical.botRadius;
	float mass = botParams.physical.mass;

	ctrl.momentOfInertia = 0.75f*mass*botRadius*botRadius;

	float alpha = frontAngle*PI/180.0f;
	float beta = backAngle*PI/180.0f;

	ctrl.theta[0] = alpha;
	ctrl.theta[1] = PI-alpha;
	ctrl.theta[2] = PI+beta;
	ctrl.theta[3] = 2*PI-beta;

	for(uint8_t i = 0; i < 4; i++)
	{
		// construct matrix for XYW velocity to motor velocity conversion
		MAT_ELEMENT(ctrl.matXYW2Motor, i, 0) = -sinf(ctrl.theta[i]);
		MAT_ELEMENT(ctrl.matXYW2Motor, i, 1) = cosf(ctrl.theta[i]);
		MAT_ELEMENT(ctrl.matXYW2Motor, i, 2) = botRadius;

		// construct matrix for motor forces to XYW forces/moment conversion
		MAT_ELEMENT(ctrl.matForceMotor2XYW, 0, i) = -sinf(ctrl.theta[i]);
		MAT_ELEMENT(ctrl.matForceMotor2XYW, 1, i) = cosf(ctrl.theta[i]);
		MAT_ELEMENT(ctrl.matForceMotor2XYW, 2, i) = botRadius;

		// construct vectors from center of mass to wheels
		MAT_ELEMENT(matCM2Wheel, i, 0) = cosf(ctrl.theta[i])*botParams.physical.botRadius - centerOfMass[0];
		MAT_ELEMENT(matCM2Wheel, i, 1) = sinf(ctrl.theta[i])*botParams.physical.botRadius - centerOfMass[1];
		MAT_ELEMENT(matCM2Wheel, i, 2) = -centerOfMass[2];
	}

	arm_mat_pinv(&ctrl.matXYW2Motor, &ctrl.matMotor2XYW);
	arm_mat_pinv(&ctrl.matForceMotor2XYW, &ctrl.matForceXYW2Motor);

	// construct matrix to calculate wheel down forces from XY acceleration
	arm_matrix_instance_f32 matQ = {3, 4, (float[12]){}};
	arm_mat_pinv(&matCM2Wheel, &matQ);

	for(uint8_t i = 0; i < 4; i++)
	{
		MAT_ELEMENT(ctrl.matForceXYZ2DownForce, i, 0) = MAT_ELEMENT(matCM2Wheel, i, 2)*MAT_ELEMENT(matQ, 0, i);
		MAT_ELEMENT(ctrl.matForceXYZ2DownForce, i, 1) = MAT_ELEMENT(matCM2Wheel, i, 2)*MAT_ELEMENT(matQ, 1, i);
		MAT_ELEMENT(ctrl.matForceXYZ2DownForce, i, 2) = MAT_ELEMENT(matCM2Wheel, i, 2)*MAT_ELEMENT(matQ, 2, i);
	}

	// Construct motor null vector
	// this simple method is only valid for robots with an equal angle between the side wheels
	pMotorNullDir[0] = MAT_ELEMENT(ctrl.matMotor2XYW, 1, 2);
	pMotorNullDir[1] = MAT_ELEMENT(ctrl.matMotor2XYW, 1, 3);
	pMotorNullDir[2] = MAT_ELEMENT(ctrl.matMotor2XYW, 1, 1);
	pMotorNullDir[3] = MAT_ELEMENT(ctrl.matMotor2XYW, 1, 0);
	float length;
	arm_power_f32(pMotorNullDir, 4, &length);
	length = sqrtf(length);
	arm_scale_f32(pMotorNullDir, 1/length, pMotorNullDir, 4);
}

void CtrlSetController(CtrlInstance* pInstance)
{
	ctrl.pNewInstance = pInstance;
}

void CtrlSetMotorsOn(uint8_t enable)
{
	ctrl.motorsOn = enable;
}

void CtrlSetLogRate(uint32_t rate)
{
	uint32_t newInterval = 1000000/rate;
	if(newInterval != ctrl.logInterval)
		ctrl.logInterval = newInterval;
}

static void stateWatchdog()
{
	if(ctrl.input.acc.updated)
		ctrl.accZ = LagElementPT1Process(&ctrl.accZLag, ctrl.input.acc.linAcc[2]);

	if(ctrl.input.gyr.updated)
		ctrl.gyrW = LagElementPT1Process(&ctrl.gyrWLag, ctrl.input.gyr.rotVel[2]);

	if(ctrl.accZ < 6.0f || fabsf(ctrl.gyrW) > 30.0f)
	{
		ctrl.misc.features &= ~(SYSTEM_MATCH_FEEDBACK_FEATURE_MOVE);

		CtrlImplBuzzerPlay(CTRL_BUZZ_BEEP_FAST);
	}

	if(ctrl.accZ > 6.5f && fabsf(ctrl.gyrW) < 25.0f)
	{
		if(!ctrl.input.power.batEmpty && (ctrl.misc.features & SYSTEM_MATCH_FEEDBACK_FEATURE_MOVE) == 0)
		{
			ctrl.misc.features |= SYSTEM_MATCH_FEEDBACK_FEATURE_MOVE;
			CtrlImplBuzzerPlay(CTRL_BUZZ_UP20);
		}
	}
}

static void doControl()
{
	static float lastExecTime;

	float now = SysTime();
	float execTime = now - lastExecTime;
	lastExecTime = now;

	if(chVTTimeElapsedSinceX(ctrl.lastDefaultPosStartTime) < MS2ST(1000))
		ctrl.input.vision.updated = 1;

	if(ctrl.pInstance && ctrl.pInstance->pControllerFunc)
	{
		// check controller execution time violation
		if(execTime > 2*ctrl.pInstance->deltaT && ctrl.buzzOnTimeViolation)
			CtrlImplBuzzerPlay(CTRL_BUZZ_UP20);

		// input timestamp is used for execution time analysis
		ctrl.input.header.timestamp = SysTimeUSec();

		// execute state estimation
		(*ctrl.pInstance->pStateEstimationFunc)(&ctrl.input, &ctrl.output);

		// call skill
		SkillsExecute();

		// assign limits
		memcpy(&ctrl.limits, &skills.output.limits, sizeof(DriveLimits));

		// call controller
		if(ctrl.pInstance->pInput && skills.output.updated)
			(*ctrl.pInstance->pInput)(skills.output.mode, skills.output.xyw);

		// execute the controller
		(*ctrl.pInstance->pControllerFunc)(&ctrl.input, &ctrl.output);

		// if motors are off or we don't have the move feature disable all motors
		if(ctrl.motorsOn == 0 || (ctrl.misc.features & SYSTEM_MATCH_FEEDBACK_FEATURE_MOVE) == 0)
			ctrl.output.ctrlMode = MOT_CTRL_OFF;

		// transfer controller output to motors
		CtrlImplApplyOutput(&ctrl.output);

		// this is the second part of execution time analysis
		ctrl.output.header.timestamp = SysTimeUSec();

		// State logging
		ctrl.state.header.timestamp = SysTimeUSec();

		if(ctrl.state.posUpdated)
			ctrl.feedbackStateUpdated[0] = 1;
		if(ctrl.state.velUpdated)
			ctrl.feedbackStateUpdated[1] = 1;
	}
}

static void doLogging()
{
	if(ctrl.botLogging.ctrlCommon)
	{
		CtrlImplWriteLogMsg(&ctrl.input);
		CtrlImplWriteLogMsg(&ctrl.state);
		CtrlImplWriteLogMsg(&ctrl.output);

		if(SysTimeUSec() - ctrl.lastKdLogTime > 5000)
		{
			ctrl.lastKdLogTime = SysTimeUSec();
			CtrlImplWriteLogMsg(&ctrl.kd);
		}

		if(SysTimeUSec() - ctrl.lastParamsLogTime > 2000000)
		{
			ctrl.lastParamsLogTime = SysTimeUSec();
			CtrlImplWriteLogMsg(&ctrl.logCfgDriveTrain);
		}
	}
}

static void clearUpdateFlags()
{
	ctrl.input.gyr.updated = 0;
	ctrl.input.acc.updated = 0;
	ctrl.input.vision.updated = 0;
	ctrl.state.posUpdated = 0;
	ctrl.state.velUpdated = 0;
}

static void checkControllerChange()
{
	if(ctrl.input.power.batEmpty)
	{
		ctrl.misc.features &= 0xFFF0;
		CtrlSetController(0);

		CtrlImplBuzzerPlay(CTRL_BUZZ_BEEP_FAST);
		CtrlImplKickerFire(IC_KD_FIRE_MODE_DISARM, 0, 0);
	}

	if(ctrl.pInstance != ctrl.pNewInstance)
	{
		// controller change requested

		if(ctrl.pInstance)
		{
			memset(ctrl.output.motorVel, 0, 4*sizeof(float));
			memset(ctrl.output.motorVoltage, 0, 4*sizeof(float));
			CtrlImplApplyOutput(&ctrl.output);

			if(ctrl.pInstance->pExit)
				(*ctrl.pInstance->pExit)();
		}

		ctrl.pInstance = ctrl.pNewInstance;

		if(ctrl.pInstance)
		{
			ctrl.ctrlIntervalUSec = (systime_t)((ctrl.pInstance->deltaT-0.0001f)*1e6f);

			LagElementPT1Init(&ctrl.accZLag, 1, 0.5f, ctrl.pInstance->deltaT);
			ctrl.accZLag.state[0] = 20.0f;
			ctrl.accZLag.state[1] = 20.0f;
			ctrl.accZLag.state[2] = 20.0f;
			ctrl.accZLag.state[3] = 20.0f;
			LagElementPT1Init(&ctrl.gyrWLag, 1, 0.5f, ctrl.pInstance->deltaT);

			ctrl.lastDefaultPosStartTime = chVTGetSystemTimeX();

			if(ctrl.pInstance->pInit)
				(*ctrl.pInstance->pInit)();
		}
		else
		{
			ctrl.output.ctrlMode = MOT_CTRL_OFF;
			CtrlImplApplyOutput(&ctrl.output);
		}
	}
}

static void sendMatchFeedback()
{
	// send status
	if(SysTimeUSec() - ctrl.lastLogTime < ctrl.logInterval)
		return;

	ctrl.lastLogTime = SysTimeUSec();

	SystemMatchFeedback feedback;
	memset(&feedback, 0, sizeof(SystemMatchFeedback));

	feedback.batteryLevel = (uint16_t)(ctrl.input.power.voltage*1000.0f);
	feedback.kickerLevel = ctrl.kd.kickerLevel;
	feedback.barrierKickCounter = (ctrl.kd.barrierInterrupted << 7) | (ctrl.kd.kickCounter & 0x7F);
	feedback.dribblerSpeed = ctrl.kd.dribblerSpeed;
	feedback.dribblerTemp = (uint8_t)(ctrl.kd.dribblerTemp*2.0f);
	feedback.hardwareId = ctrl.misc.hardwareId;
	feedback.features = ctrl.misc.features;

	if(ctrl.kd.dribblerOverheated)
	{
		feedback.features &= ~SYSTEM_MATCH_FEEDBACK_FEATURE_DRIBBLE;
	}

	if(ctrl.feedbackStateUpdated[0])
	{
		ctrl.feedbackStateUpdated[0] = 0;

		feedback.curPosition[0] = (int16_t)(ctrl.state.pos[0]*1000.0f);
		feedback.curPosition[1] = (int16_t)(ctrl.state.pos[1]*1000.0f);
		feedback.curPosition[2] = (int16_t)(ctrl.state.pos[2]*1000.0f);
	}
	else
	{
		feedback.curPosition[0] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curPosition[1] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curPosition[2] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
	}

	if(ctrl.feedbackStateUpdated[1])
	{
		ctrl.feedbackStateUpdated[1] = 0;
		feedback.curVelocity[0] = (int16_t)(ctrl.state.vel[0]*1000.0f);
		feedback.curVelocity[1] = (int16_t)(ctrl.state.vel[1]*1000.0f);
		feedback.curVelocity[2] = (int16_t)(ctrl.state.vel[2]*1000.0f);
	}
	else
	{
		feedback.curVelocity[0] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curVelocity[1] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curVelocity[2] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
	}

	PacketHeader header;
	header.section = SECTION_SYSTEM;
	header.cmd = CMD_SYSTEM_MATCH_FEEDBACK;
	CtrlImplSendNetworkPacket(&header, (uint8_t*) &feedback, sizeof(SystemMatchFeedback));
}

static void updateParameters()
{
	ctrl.logCfgDriveTrain.motor2WheelRatio = botParams.driveTrain.motor2WheelRatio;
	ctrl.logCfgDriveTrain.Km = botParams.driveTrain.motor.Km;
	ctrl.logCfgDriveTrain.Kn = botParams.driveTrain.motor.Kn;
	ctrl.logCfgDriveTrain.Ke = botParams.driveTrain.motor.Ke;
	ctrl.logCfgDriveTrain.Kf = botParams.driveTrain.motor.Kf;
	ctrl.logCfgDriveTrain.R = botParams.driveTrain.motor.R;
	ctrl.logCfgDriveTrain.L = botParams.driveTrain.motor.L;

	ctrl.logCfgDriveTrain.header.timestamp = SysTimeUSec();
}
