/*
 * ctrl_motor.c
 *
 *  Created on: 08.03.2013
 *      Author: AndreR
 */

#include <main/ctrl_motor.h>
#include "struct_ids.h"
#include "util/angle_math.h"
#include "util/arm_mat_util_f32.h"

#include <string.h>

static void init();
static void controller(CtrlInput* pInput, CtrlOutput* pOutput);
static void stateEstimation(CtrlInput* pInput, const CtrlOutput* pOutput);
static void input(uint8_t mode, float* pCmd);

static const ConfigFileDesc configFileDescVel =
	{ SID_CFG_CTRL_MOTOR_VEL, 0, "ctrl_motor/vel", 3, (ElementDesc[]) {
		{ FLOAT, "max_vel_xy", "", "maxVel/XY" },
		{ FLOAT, "max_vel_w", "", "maxVel/W" },
		{ UINT8, "override_bat", "", "Override Bat." },
	} };

CtrlMotor ctrlMotor = {
	.configVel = {
		.maxXY = 1.0f,
		.maxW = 3.0f,
		.overrideBat = 1.0f,
	},
};
CtrlInstance ctrlMotorInstance = {
	.pInit = &init,
	.pExit = 0,
	.pStateEstimationFunc = &stateEstimation,
	.pControllerFunc = &controller,
	.pInput = &input,
	.deltaT = 0.001f };

static void init()
{
	ctrlMotor.pConfigFileVel = ConfigOpenOrCreate(&configFileDescVel, &ctrlMotor.configVel, sizeof(CtrlMotorConfigVel), 0);

	memset(ctrlMotor.velSetpoint, 0, sizeof(float)*4);

	ctrlMotor.matCurMotor.pData = ctrlMotor.curWheel;
	ctrlMotor.matCurMotor.numCols = 1;
	ctrlMotor.matCurMotor.numRows = 4;

	ctrlMotor.matVelSetpoint.pData = ctrlMotor.velSetpoint;
	ctrlMotor.matVelSetpoint.numCols = 1;
	ctrlMotor.matVelSetpoint.numRows = 4;
}

static void stateEstimation(CtrlInput* pInput, const CtrlOutput* pOutput)
{
	(void)pOutput;

	// Encoders
	memcpy(ctrlMotor.curWheel, pInput->enc.vel, sizeof(float)*4);

	for(uint8_t i = 0; i < 4; i++)
		ctrlMotor.curWheel[i] *= botParams.physical.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;	// value is now speed over ground [m/s]

	// velocity state update
	float localVel[3];
	arm_matrix_instance_f32 matVel = {3, 1, localVel};
	arm_mat_mult_f32(&ctrl.matMotor2XYW, &ctrlMotor.matCurMotor, &matVel);

	float globalVel[2];
	CtrlUtilTurnLocal2Global(ctrl.state.pos[2], localVel[0], localVel[1], globalVel, globalVel+1);
	ctrl.state.vel[0] = globalVel[0];
	ctrl.state.vel[1] = globalVel[1];
	ctrl.state.vel[2] = localVel[2];
	ctrl.state.velUpdated = 1;

	if(pInput->vision.updated)
	{
		// position update (feed-through)
		memcpy(ctrl.state.pos, pInput->vision.pos, sizeof(float)*3);
		ctrl.state.posUpdated = 1;
	}
	else
	{
		// integrate encoders if no pos update is available
		ctrl.state.pos[0] += ctrl.state.vel[0]*ctrlMotorInstance.deltaT;
		ctrl.state.pos[1] += ctrl.state.vel[1]*ctrlMotorInstance.deltaT;
		float newOrient = AngleNormalize(ctrl.state.pos[2] + ctrl.state.vel[2]*ctrlMotorInstance.deltaT);
		ctrl.state.pos[2] = newOrient;
		ctrl.state.posUpdated = 1;
	}
}

static void controller(CtrlInput* pInput, CtrlOutput* pOutput)
{
	(void)pOutput;
	(void)pInput;

	// convert speed over ground to motor output
	for(uint8_t i = 0; i < 4; i++)
		pOutput->motorVel[i] = ctrlMotor.velSetpoint[i]*CTRL_WHEEL_TO_MOTOR_RATIO * 1/botParams.physical.wheelRadius;

	pOutput->ctrlMode = MOT_CTRL_VELOCITY;
}

static void input(uint8_t mode, float* pCmd)
{
	switch(mode)
	{
		case CTRL_MODE_LOCAL_VEL:
		{
			CtrlMotorSetVelTarget(pCmd[0], pCmd[1], pCmd[2]);
		}
		break;
		case CTRL_MODE_GLOBAL_POS:
		{
		}
		break;
	}
}

void CtrlMotorSetVelTarget(float x, float y, float w)
{
	float xyw[3] = {x, y , w};
	arm_matrix_instance_f32 matXYW = {3, 1, xyw};

	arm_mat_mult_f32(&ctrl.matXYW2Motor, &matXYW, &ctrlMotor.matVelSetpoint);
}
